/***************************************************************************
                          command.cpp  -  description
                             -------------------
    begin                : Sun Aug 5 2001
    author               : Dmitri Dmitrienko
    copyright            : (C) 2001, 2007, Dmitri Dmitrienko
    www                  : http://dd.cron.ru
    license              : This source file is subject to version 1.01 of 
                           the License,  that is bundled with this package 
                           in the file LICENSE, and is available at through 
                           the world-wide-web at http://dd.cron.ru/license
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef WIN32
#include "config.w32.h"
#if defined(_DEBUG)
#define _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#include <crtdbg.h>
#endif
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#if HAVE_UNISTD_H
#include <unistd.h>
#endif


#include <stdio.h>
#include <string.h>

#include <stdlib.h>
#include <signal.h>
#include <setjmp.h>

#include "dbgbase/dbgbase_classes.h"
#include "dbgbase/dbgbase_globals.h"
#include "dbgbase/dbg_net.h"

#include "dbgclient_classes.h"
#include "command.h"

#ifdef WIN32
#include <direct.h>
#endif

#include <readline/readline.h>
#include <readline/history.h>

#define msg_dbg_already "Debug session is active. Can not have more than one at a time.\n"

static CommandEntry info_commands[] = {
	{"files", false, htNoArg, NULL, handler_info_sources, "\
Source files in the program being debugged."},
	{"sources", false, htNoArg, NULL, handler_info_sources, "\
Source files in the program being debugged."},
	{"source", false, htNoArg, NULL, handler_info_source, "\
Information about the current source file."},
	{"program", false, htNoArg, NULL, handler_info_program, "\
Shows execution status of the program."},
	{"breakpoints", true, htNoneOrManyUShortArg, NULL, handler_info_breakpoints, "\
Status of user-settable breakpoints, or breakpoint number NUMBER.\n\
The \"Type\" column indicates one of:\n\
        breakpoint     - normal breakpoint\n\
The \"Disp\" column contains one of \"keep\", \"del\", or \"dis\" to indicate\n\
the disposition of the breakpoint after it gets hit.  \"dis\" means that the\n\
breakpoint will be disabled.  The \"What\" column indicate the\n\
file and line number."},
	ENDCOMMAND
};

static CommandList info_cmdlist(info_commands);

static CommandEntry set_commands[] = {
	{"port", false, htOneUShortArg, NULL, handler_set_port, "\
Set TCP/IP port to be used by the dbg listener for incoming connections\n\
from the server part. By default its vaule is 10001"},
	{"mode", false, htOneStrArg, NULL, handler_set_mode, "\
Set HTTP debugging mode. Otherwise CGI mode will be used."},
	{"cgiexec", false, htOneStrArg, NULL, handler_set_cgiexec, "\
Set CGI executable to be used as a CGI HANDLER in local debugging mode."},
	{"mapurlroot", false, htOneStrArg, NULL, handler_set_mapurlroot, "\
Set URL root to be used for mapping script files to and from URLs."},
	{"maplocalroot", false, htOneStrArg, NULL, handler_set_maplocalroot, "\
Set local root to be used for mapping script files to and from remote files."},
	{"mapremoteroot", false, htOneStrArg, NULL, handler_set_mapremoteroot, "\
Set remote root to be used for mapping remote script files to and from local files."},
	ENDCOMMAND
};

static CommandList set_cmdlist(set_commands);

static CommandEntry show_commands[] = {
	{"port", false, htNoArg, NULL, handler_show_port, "\
Show TCP/IP port to be used by the dbg listener for incoming connections\n\
from the server part. By default its value is 10001"},
	{"mode", false, htNoArg, NULL, handler_show_mode, "\
Show debugging mode. Off means CGI mode while on - HTTP."},
	{"cgiexec", false, htNoArg, NULL, handler_show_cgiexec, "\
Show CGI program to be used as a CGI HANDLER to execute a script."},
	{"mapurlroot", false, htNoArg, NULL, handler_show_mapurlroot, "\
Show URL root to be used for script files<->URLs mapping."},
	{"maplocalroot", false, htNoArg, NULL, handler_show_maplocalroot, "\
Show local root to be used for mapping script files to and from remote files."},
	{"mapremoteroot", false, htNoArg, NULL, handler_show_mapremoteroot, "\
Show remote root to be used for mapping remote script files to and from local files."},
	ENDCOMMAND
};

static CommandList show_cmdlist(show_commands);

static CommandEntry main_commands[] = {
// SETTINGS
	{"set", false,  htOneStrArg, &set_cmdlist, NULL, "\
set port -- Set TCP/IP port number.\n\
set mode -- Set HTTP mode of debugging.\n\
set cgiexec -- Set CGI executable.\n\
set mapurlroot -- Set URL root for mapping.\n\
set maplocalroot -- Set local root for mapping.\n\
set mapremoteroot -- Set remote root for mapping."},
	{"show", false,  htOneStrArg, &show_cmdlist, NULL, "\
Common command for showing settings.\n\
show port -- Show TCP/IP port number.\n\
show mode -- Show mode of debugging.\n\
show cgiexec -- Show CGI executable to be used in local debugging mode.\n\
show mapurlroot -- Show URL root for mapping.\n\
show maplocalroot -- Show local root for mapping.\n\
show mapremoteroot -- Show remote root for mapping."},
// SESSION
	{"file", false,  htOneStrArg, NULL, handler_file, "\
Use FILE as program (script) to be debugged.\n\
It is the program executed when you use the `run' command.\n\
If FILE cannot be found as specified, your execution directory path\n\
($PATH) is searched for a command of that name."},
	{"run",  false, htAnyArgs, NULL, handler_run, "\
Start debugged program (script). An URL or a path to the script has to be provided."},
	{"kill", true,  htNoArg, NULL, handler_kill, "\
Kill execution of program being debugged."},
	{"quit", false, htNoArg, NULL, handler_quit, "\
Exit dbg."},
	{"q", false, htNoArg, NULL, handler_quit, "\
Exit dbg."},
	{"listen",  false, htOneOrNoneShortArg, NULL, handler_listen, "\
Listen on specified port and wait for debug session.\n\
Optional argument specifies number of seconds to wait."},
// STEPPING
	{"step", true,  htNoArg, NULL, handler_step, "\
Step program until it reaches a different source line."},
	{"next", true,  htNoArg, NULL, handler_next, "\
Step program, proceeding through subroutine calls.\n\
Like the \"step\" command as long as subroutine calls do not happen;\n\
when they do, the call is treated as one instruction."},
	{"finish", true,  htNoArg, NULL, handler_finish, "\
Execute until selected stack frame returns.\n\
Upon return, the value returned is printed and put in the value history."},
	{"cont", true,  htNoArg, NULL, handler_cont, "\
Continue program being debugged, after error or breakpoint.\n\
If proceeding from breakpoint, a number N may be used as an argument,\n\
which means to set the ignore count of that breakpoint to N - 1 (so that\n\
the breakpoint won't break until the Nth time it is reached)."},
// BACK TRACE
	{"up", true,  htOneOrNoneShortArg, NULL, handler_up, "\
Select and print stack frame that called this one.\n\
An argument says how many frames up to go."},
	{"down", true,  htOneOrNoneShortArg, NULL, handler_down, "\
Select and print stack frame called by this one.\n\
An argument says how many frames down to go."},
	{"where", true,  htOneOrNoneShortArg, NULL, handler_where, "\
Print backtrace of all stack frames, or innermost COUNT frames.\n\
With a negative argument, print outermost -COUNT frames.\n\
Use of the 'full' qualifier also prints the values of the local variables."},
	{"bt", true,  htOneOrNoneShortArg, NULL, handler_where, "\
Print backtrace of all stack frames, or innermost COUNT frames.\n\
With a negative argument, print outermost -COUNT frames.\n\
Use of the 'full' qualifier also prints the values of the local variables."},
	{"frame", true,  htOneOrNoneShortArg, NULL, handler_frame, "\
Select and print a stack frame.\n\
With no argument, print the selected stack frame.  (See also \"info frame\").\n\
An argument specifies the frame to select.\n\
It can be a stack frame number or the address of the frame.\n\
With argument, nothing is printed if input is coming from\n\
a command file or a user-defined command."},
// BREAKPOINTS
	{"break", true, htOneStrArg, NULL, handler_break, "\
Set breakpoint at specified line.\n\
If line number is specified, break at start of code for that line.\n\
With no arg, uses current execution address.\n\
Multiple breakpoints at one place are permitted, and useful if conditional."},
	{"tbreak", true, htOneStrArg, NULL, handler_tbreak, "\
Set a temporary breakpoint.  Args like \"break\" command.\n\
Like \"break\" except the breakpoint is only temporary,\n\
so it will be deleted when hit."},
	{"enable", true, htOneUShortArg, NULL, handler_enable, "\
Enable breakpoint."},
	{"disable", true, htOneUShortArg, NULL, handler_disable, "\
Disable breakpoint."},
	{"delete", true, htOneUShortArg, NULL, handler_delete, "\
Delete breakpoint."},
	{"condition", true, htAnyArgs, NULL, handler_condtion, "\
Set a condition for breakpoint."},
// EVALUATION
	{"display", false,  htOneOrNoneStrArg, NULL, handler_display, "\
Print value of expression each time the program stops."},
	{"undisplay", true,  htNoneOrManyUShortArg, NULL, handler_undisplay, "\
Stop printing value(s) each time the program stops."},
	{"print", true,  htOneStrArg, NULL, handler_print, "\
Print value of expression."},
	{"output", false, htNoArg, NULL, handler_output, "\
Prints out all the output produced by the program being debugged."},
// INFORMATION
	{"help", false, htAnyArgs, NULL, handler_help, "\
Print list of commands."},
	{"info", false, htNoArg, &info_cmdlist, NULL, "\
Generic command for showing things about the program being debugged."},
	{"pwd", false,  htNoArg, NULL, handler_pwd, "\
Print working directory.  This is used for your program as well."},
	{"list", false,  htAnyArgs, NULL, handler_list, "\
List current source context being debugged. Use \"list auto\" command\n\
to have it printed each time the program is breaked."},
	ENDCOMMAND
};

static CommandList main_cmdlist(main_commands);

////////////////       //////////////////
// 				GLOBALS           	   //
////////////////       //////////////////

// Config parameters
#ifdef WIN32
static std::string cgihandler = "php.exe";
#else
static std::string cgihandler = "php";
#endif
static int port = 0;
static bool dbg_HTTP_mode = false;
static filemapper mapper(
	dbg_HTTP_mode, 
	"http://localhost/",
	"/usr/local/apache/htdocs/",
	"/usr/local/apache/htdocs/");

// Run-time parameters
static std::string main_script;
static std::string run_params;

// derived classes instances
static Listener listener;
static DbgSite *fsite = NULL;

// flags
static SessionState sess_state = ssNotStarted;
static bool run_request = false;
static bool quit_request = false;
static bool async_pause = false;
static bool async_term = false;
static bool dbg_AUTO_LIST = false;

// misc
#define DBG_PROMPT "dbg>"

// log
static FILE* log = NULL;

#ifdef WIN32
static char* logfile = "dbg.log";
#else
static char* logfile = "/tmp/dbg.log";
#endif

static BP_ACTION UIAction;
#define LOG(a) \
	if (log) fprintf(log, a)

#define LOGLN(a) \
	if (log) fprintf(log, "%s\n", a)


// Backtrace
static dbgint bt_curr_idx = 0;
static dbgint bt_curr_scopeid = CURLOC_SCOPE_ID;

// Watches
typedef struct tag_watch_rec {
	int idx;
	std::string eval_str;
} watch_rec;
typedef std::vector<watch_rec> watch_list;
static dbgint last_watch_idx = 0;
static watch_list watches;
static dbgint print_counter = 0;

#if defined(WIN32) && defined(_DEBUG)
#define MEMCHK_ON _CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF)
#define MEMCHK_OFF _CrtSetDbgFlag(0)
#else
#define MEMCHK_ON 
#define MEMCHK_OFF
#endif




// CommandList

CommandEntry *CommandList::lookup_cmd(const char *cmd) {
	int L, H, i, cmp;
	iterator entry;
	L = 0;
	H = end() - begin() - 1;
	while (L <= H) {
		i = (L + H) >> 1;
		entry = (iterator)begin();
		entry += i;
		cmp = safestricmp(entry->name, cmd);
		if (cmp < 0) {
			L = i + 1;
		}
		else {
			if (cmp == 0) return (&*entry);
			H = i - 1;
		}
	}
	return NULL;
}


CmdProcResult chkShortArg(char *argv, bool is_signed) {
	long v;
	char *p;
	v = strtol (argv, &p, 0);
	if (p[0] != '\0') {
		fprintf(rl_outstream, "Invalid value supplied `%s'.\n",p);
		return cprOutRangeVal;
	}
	if (is_signed) {
		if (v < 0 || v > 65535) {
			fprintf(rl_outstream, "Out of range value supplied.\n");
			return cprOutRangeVal;
		}
	} else {
		if (v <= -32768 || v > 32767) {
			fprintf(rl_outstream, "Out of range value supplied.\n");
			return cprOutRangeVal;
		}
	}
	return cprOk;
}

// process_command

static CmdProcResult process_command(SessionState st, char *cmd_line) {
	char *pstart;
	char *pend;
	char *tok;
	CommandEntry *cmd;
	SimpleStringArray ssa;
	pstart = cmd_line;
	while (*pstart == ' ' || *pstart == '\t') pstart++;
	pend = cmd_line + strlen(cmd_line) - 1;
	while (pend >= pstart && (*pend == ' ' || *pend == '\t')) pend--;
	if (pend < pstart) return cprOkEmpty;
	if (pstart[0] == '#') return cprOkIgnored;

	CommandList* pcmdlist = &main_cmdlist;
	
// find cmd handler
#define seps " \t"
	char *prevcmd=NULL;
	while (1) {
		tok = strtok(pstart, seps);
		pstart = NULL;
		if (!tok) {
			if (pcmdlist != &main_cmdlist) {
				fprintf(rl_outstream, "\"%s\" must be followed with the name of subcommand.\n",prevcmd);
			}
			return cprOkEmpty;
		}
		prevcmd = tok;

		cmd = pcmdlist->lookup_cmd(tok);
		if (!cmd) {
			if (pcmdlist == &main_cmdlist) {
				fprintf(rl_outstream, "Undefined command: \"%s\".  Try \"help\".\n", tok);
			} else {
				fprintf(rl_outstream, "Undefined subcommand: \"%s\".  Try \"help\".\n", tok);
			}
			return cprUnknownCmd;
		}
		
		if (cmd->reqSessStarted && (st != ssStarted)) {
		        fprintf(rl_outstream, "The program is not being run.\n");
			return cprOkIgnored;
		}

		if (cmd->subCommandList) {
			pcmdlist = cmd->subCommandList;
		} else {
			break;
		}
	}

	while (1) {
		tok = strtok(NULL, seps);
		if (!tok) break;
		ssa.add(tok);
	}	
	

	int argc = ssa.count();
	switch (cmd->ht) {
		case htNoArg:
			if (argc != 0) fprintf(rl_outstream, "No arguments expected. All are ignored\n");
			break;
		case htOneStrArg:
		case htOneUShortArg:
			if (argc < 1) {
				fprintf(rl_outstream, "One argument expected.\n");
				return cprNoArg;
			}
		// FOLLOW!
		case htOneOrNoneStrArg:
		case htOneOrNoneUShortArg:
		case htOneOrNoneShortArg:
			if (argc > 1) fprintf(rl_outstream, "Extra argument(s) supplied. Only one expected, all above are ignored\n");
			if (argc == 1 && 
			    (cmd->ht == htOneOrNoneUShortArg || 
			     cmd->ht == htOneUShortArg || 
				 cmd->ht == htOneOrNoneShortArg)) 
			{
				CmdProcResult ret_val;
				ret_val = chkShortArg(ssa[0],
					cmd->ht == htOneOrNoneUShortArg || cmd->ht == htOneUShortArg);
				if (ret_val != cprOk) return ret_val;
			}
			break;
		case htNoneOrManyUShortArg:
			{
				CmdProcResult ret_val;
				for (int i=0; i<argc; i++) {
					ret_val = chkShortArg(ssa[i], false);
					if (ret_val != cprOk) return ret_val;
				}
			}
			break;
		case htAnyArgs:
			break;
	}

	return (cmd->func) ? (cmd->func(ssa.count(), ssa.array())) : cprOkIgnored;
}

//////////////              //////////////
//               SIGNALS 			    //
//////////////______________//////////////


static void handle_sigterm(int sig) {
	signal(sig, handle_sigterm);
	async_term = true;
}

static void handle_sigio(int sig) {
	signal(sig, handle_sigio);
	async_term = true;
}

static void handle_sigint(int sig) {
	signal(sig, handle_sigint);
	async_pause = true;
#ifdef WIN32
//	rl_done = 1;
#else
//	rl_free_line_state();
//	rl_cleanup_after_signal();
#endif
}

//////////////              //////////////
//                WAITUI 			    //
//////////////______________//////////////

static void print_location(RequestorBase *requestor) {
	dbgint mod_no = 0, line_no = 0;
	std::string mod_name, scope_name, warn;
	if (SUCCEEDED(requestor->getcurloc(&mod_no, &line_no, &scope_name)) && 
		SUCCEEDED(requestor->requestdocfilename(mod_no, &mod_name))) {
		if (!mapper.mapfilename(mkRemoteToLocal, mod_name, mod_name, warn)) {
			fprintf(rl_outstream, "%s\n", warn.c_str());
		}
		fprintf(rl_outstream, "%s:%ld <%s>\n", mod_name.c_str(), line_no, scope_name.c_str());
	} else {
		fprintf(rl_outstream, "Can not get current location.\n");
	}
}

static void process_line(char *line) {
//	if (async_break) return;

	signal(SIGTERM, handle_sigterm);
#if defined(SIGIO)
	signal(SIGIO, handle_sigio);
#endif
	signal(SIGINT, handle_sigint);
#if defined(SIGPIPE) && defined(SIG_IGN)
	signal(SIGPIPE, SIG_IGN);
#endif

	CmdProcResult rslt;
	if (line==NULL || strlen(line) == 0) return;
	LOGLN(line);
	MEMCHK_OFF;
	add_history(line);
	MEMCHK_ON;
	rslt = process_command(sess_state, line);
	if (rslt == cprOkQuit) quit_request = true;
}

void dodisplaywatch(int idx, bool showerr);

BP_ACTION WaitUI(RequestorBase *requestor, BR_REASON reason) {
	UIAction = (BP_ACTION)-1;
	char *aline;

	DBG_TRACE(("WaitUI\n"));
	bt_curr_idx = 0;
	bt_curr_scopeid = CURLOC_SCOPE_ID;

	if (reason != BR_END_SESSION) {
		print_location(requestor);
		fsite->update_curr_src();
		dodisplaywatch(-1,false);
		if (dbg_AUTO_LIST) {
		    handler_list(0, NULL);
		}
	}
	while (!quit_request && UIAction == (BP_ACTION)-1) {
		MEMCHK_OFF;
		aline = readline(DBG_PROMPT);
		MEMCHK_ON;
		
		process_line(aline);
		free(aline);
		if (run_request) {
			UIAction = BP_ABORT;
			break;
		}
	}
	async_term = false;
	async_pause = false;
	if (quit_request) UIAction = BP_ABORT;
	return UIAction;
}

//////////////              //////////////
//                DBGSITE 			    //
//////////////______________//////////////

HRESULT DbgSite::isalive() {
	if (async_term) {
	    fprintf(rl_outstream, "terminate request.\n");
	    async_term = false;
	    return E_FAIL;
	}
	if (async_pause) {
	    fprintf(rl_outstream, "pause request.\n");
	    async_pause = false;
	    return S_FALSE;
	}
	return S_OK;
}

HRESULT DbgSite::initialize(RequestorBase *req, int dbg_ver, char *dbg_descr) {
	// Assign globals
	frequestor = req;
	foutput = "";
	fcurr_src = "";
	fcurr_mod_no = 0;
	fcurr_line_no = 0;
	fcurr_mod_name = "";
	sess_state = ssStarted;
	last_watch_idx = 0;
	print_counter = 0;
	watches.clear();
	async_term = false;
	async_pause = false;
	fprintf(rl_outstream, "Starting program: `%s'\n", main_script.c_str());
	fsite = this;
	return S_OK;
}

HRESULT DbgSite::handleerror(int errtype, char *errmsg, BP_ACTION *action) {
	if (errmsg) 
	    fprintf(rl_outstream, "%s\n", errmsg);
	*action = WaitUI(requestor(), BR_UNKNOWN);
	return S_OK;
}

HRESULT DbgSite::handlebreakpoint(BR_REASON reason, BP_ACTION *action) {
	DBG_TRACE(("DbgSite::handlebreakpoint\n"));
	if (reason == BR_START_SESSION) {
		*action = BP_STEP_INTO;
		return S_OK;
	}
	if (reason == BR_END_SESSION) {
		fprintf(rl_outstream, "End of session pended.\n");
	}
	*action = WaitUI(requestor(), reason);
	return S_OK;
}

HRESULT DbgSite::log(int mod_no, int line_no, char *mod_name, int type, char *logmsg, int loglen) {
	if (type == LT_OUTPUT) {
	    DBG_TRACE(("OUTPUT ADDED, \"%s\" (len=%d).\n", logmsg, loglen));
	    foutput.append(logmsg, loglen);
	}
	return S_OK;
}

HRESULT DbgSite::close() {
	fsite = NULL;
	sess_state = ssStopped;
	if (!run_request)
		fprintf(rl_outstream, "\nProgram exited.\n");		
	return S_OK;
}

void DbgSite::update_curr_src() {
	dbgint mod_no = 0, line_no = 0;
	std::string scope_name, warn;
	if (SUCCEEDED(frequestor->getcurloc(&mod_no, &line_no, &scope_name))) {
		fcurr_line_no = line_no;
		if (fcurr_mod_no == mod_no)
		    return;
		fcurr_mod_no = mod_no;
		if (!frequestor->requestdocfilename(mod_no, &fcurr_mod_name) ||
		    !mapper.mapfilename(mkRemoteToLocal, fcurr_mod_name, fcurr_mod_name, warn) ||
		    !fileexists(fcurr_mod_name) ||
		    !readfile(fcurr_mod_name, fcurr_src)) {
			frequestor->requestdocsrc(fcurr_mod_no, &warn, &fcurr_src);
		}
	} else {
		fcurr_mod_no = 0;
		fcurr_line_no = 0;
		fcurr_src = "";
		fcurr_mod_name = "";
	}
}


//////////////////////

void do_run() {
	HRESULT hr;
	SessRequest sessionreq;
	DBGS_RESULT result;
	DbgSiteBase *asite = NULL;
	
	if (fsite) {
		fprintf(rl_outstream, msg_dbg_already);
		return;
	}
	
	sessionreq.port = port;
	sessionreq.address = "";

	if (dbg_HTTP_mode) {
		std::string warn;
		sessionreq.server_type = SVR_HTTP;
		if (!mapper.mapfilename(mkLocalToURL, main_script, sessionreq.request, warn)) {
			fprintf(rl_outstream, "%s\n", warn.c_str());
			return;
		}
		sessionreq.cgi_handler="";
	} else {
		sessionreq.server_type = SVR_CGI;
		sessionreq.cgi_handler=cgihandler;
		sessionreq.request = main_script;
	}
	
	if (run_params.length() > 0) {
		if (sessionreq.request.find('?') != sessionreq.request.npos) {
			sessionreq.request.append("&");
		} else {
			sessionreq.request.append("?");
		}
		sessionreq.request.append(run_params);
	}
	sessionreq.SOF_flags = SOF_BREAKONLOAD | SOF_BREAKONFINISH | SOF_SEND_OUTPUT;
#ifdef _DEBUG
	sessionreq.timeoutms_waitsess = 1000000;
	sessionreq.timeoutms_client = 3000000;
#else
	sessionreq.timeoutms_waitsess = 10000;
	sessionreq.timeoutms_client = 30000;
#endif
	asite = new DbgSite();
	hr = listener.startsession(sessionreq, asite, &result);
	if (FAILED(hr)) {
		switch (result) {
			case DBGSE_WRONGREQ:
				fprintf(rl_outstream, "Wrong session request passed\n");
				break;
			case DBGSE_FILENOTFOUND:
				fprintf(rl_outstream, "File `%s' not found.\n",main_script.c_str());
				break;
			case DBGSE_CGIHANDLERNOTFOUND:
				fprintf(rl_outstream, "CGI handler `%s' not found.\n",cgihandler.c_str());
				break;
			case DBGSE_HTTPHOSTNOTFOUND:
				fprintf(rl_outstream, "HTTP host is not found. Please, check URL:\n\t%s\n",main_script.c_str());
				break;
			case DBGSE_TIMEOUTSTARTSESS:
				fprintf(rl_outstream, "Session start timeout has expired.\n");
				break;
			case DBGSE_FAILEDBINDSOCKET:
				fprintf(rl_outstream, "Port %d is already in-use.\n",sessionreq.port);
				break;
			default:
				fprintf(rl_outstream, "Internal error (%d).\n",(int)result);
		}
		fprintf(rl_outstream, "Program is not being run.\n");
		return;
	}
	return;
}

int count_lines(const std::string &src) {
	int ret_val = 0;
	int len = src.length();
	if (len == 0) return 0;
	int ofs = 0;
	while (true) {
	    ofs = src.find("\n", ofs);
	    ret_val++;
	    if (ofs == src.npos)
		break;
	    ofs++;
	}
	DBG_TRACE(("count_line->%d\n", ret_val));
	return ret_val;	
}

int seek_line(const std::string &src, int line_no) {
	int ret_val = src.npos;
	int ofs = 0;
	int c_line = 1;
	DBG_TRACE(("seek_line(%d)\n", line_no));
	if (c_line == line_no) 
	    return 0;
	if (line_no < 1) 
	    return src.npos;
	
	while (true) {
	    ofs = src.find("\n", ofs);
	    c_line++;
	    if (ofs == src.npos)
		break;
	    ofs++;
	    ret_val = ofs;
	    if (c_line >= line_no)
		break;
	}
	return ret_val;
}

int print_lines(FILE* fp, const std::string &src, int startline, int count, int markline) {
	int ret_val = 0;
	int ofs, eofs, len;
	std::string line;
	
	DBG_TRACE(("print_lines(%d, %d, %d)\n", startline, count, markline));
	if (startline == 0) {
	    startline++;
	    count--;
	}
	if (count <= 0)
	    return 0;
	if (startline < 0) { // start from the bottom
	    startline += count_lines(src);
	}
	ofs = seek_line(src, startline);
	len = src.length();
        DBG_TRACE(("len=%d\n", len));
	while (ofs < len && count > 0) {
	    eofs = src.find("\n", ofs);
	    if (eofs == src.npos) {
		eofs = src.length() - 1;
		if (eofs >= 0 && src[eofs] != '\n') 
		    eofs++;
	    }
	    DBG_TRACE(("ofs=%d, len=%d, s=%s\n", ofs, eofs-ofs, src.substr(ofs, eofs-ofs).c_str()));
	    line.assign(src.substr(ofs, eofs - ofs));
	    fprintf(fp, "%d%s%s\n", startline, (startline==markline) ? "> " : "| ", line.c_str());
	    count--;
	    startline++;
	    ofs = eofs + 1;
	}
}

//////////////              //////////////
//               HANDLERS 			    //
//////////////______________//////////////


// SETTINGS

CmdProcResult handler_set_port(int argc, const char *argv[]) {
	unsigned long v;
	char *p;
	v = strtoul (argv[0], &p, 0);
	port = (unsigned short)v;
	return cprOk;
}

CmdProcResult handler_show_port(int argc, const char *argv[]) {
	fprintf(rl_outstream, "%d.\n", (port)?(port):(DEFAULT_REQ_PORT));
	return cprOk;
}

CmdProcResult handler_set_mode(int argc, const char *argv[]) {
	dbg_HTTP_mode = (stricmp(argv[0], "on") == 0);
	mapper.is_HTTP = dbg_HTTP_mode;
	DBG_TRACE(("set mode->%d (%s)", dbg_HTTP_mode, argv[0]));
	return cprOk;
}

CmdProcResult handler_show_mode(int argc, const char *argv[]) {
	fprintf(rl_outstream, "HTTP mode is\n%s.\n", dbg_HTTP_mode? "on":"off");
	return cprOk;
}

CmdProcResult handler_set_cgiexec(int argc, const char *argv[]) {
	std::string fname = argv[0];
	if (!findfile(fname)) {
		fprintf(rl_outstream, "CGI handler \"%s\" not found.\n", fname.c_str());
		return cprFailed;
	}
	cgihandler = fname;
	return cprOk;
}


CmdProcResult handler_show_cgiexec(int argc, const char *argv[]) {
	fprintf(rl_outstream, "%s.\n", cgihandler.c_str());
	return cprOk;
}

CmdProcResult handler_set_mapurlroot(int argc, const char *argv[]) {
	mapper.set_url_root(argv[0]);
	return cprOk;
}

CmdProcResult handler_show_mapurlroot(int argc, const char *argv[]) {
	std::string s;
	mapper.get_url_root(s);
	fprintf(rl_outstream, "%s.\n", s.c_str());
	return cprOk;
}

CmdProcResult handler_set_maplocalroot(int argc, const char *argv[]) {
	mapper.set_local_root(argv[0]);
	return cprOk;
}

CmdProcResult handler_show_maplocalroot(int argc, const char *argv[]) {
	std::string s;
	mapper.get_local_root(s);
	fprintf(rl_outstream, "%s.\n", s.c_str());
	return cprOk;
}

CmdProcResult handler_set_mapremoteroot(int argc, const char *argv[]) {
	mapper.set_remote_root(argv[0]);
	return cprOk;
}

CmdProcResult handler_show_mapremoteroot(int argc, const char *argv[]) {
	std::string s;
	mapper.get_remote_root(s);
	fprintf(rl_outstream, "%s.\n", s.c_str());
	return cprOk;
}


// SESSION

CmdProcResult handler_file(int argc, const char *argv[]) {
	std::string fname = argv[0];
	
	if (fname.find("://") != fname.npos) { // that's URL, let's map it back to local fname
		std::string mod_name, warn;
		if (!mapper.mapfilename(mkURLToLocal, fname, mod_name, warn)) {
			fprintf(rl_outstream, "%s\n", warn.c_str());
			return cprFailed;
		} else {
			fprintf(rl_outstream, "resolved file name: %s\n", mod_name.c_str());
			fname.assign(mod_name);
		}
	}
	
	if (!findfile(fname)) {
		fprintf(rl_outstream, "\"%s\": No such file.\n", fname.c_str());
		return cprFailed;
	}
	fprintf(rl_outstream, "Reading symbols from %s...done.\n", fname.c_str());
	main_script = fname;
	return cprOk;
}

CmdProcResult handler_run(int argc, const char *argv[]) {
	if (fsite != NULL) {
		UIAction = BP_CONTINUE;
		return cprOk;
	}

	run_request = false;	
	if (main_script.length() == 0) {
		fprintf(rl_outstream, "No symbol table is loaded.  Use the \"file\" command.\n");
		return cprFailed;
	}
	run_request = true;	
	run_params = "";
	for (int i=0; i<argc; i++) {
		if (i<argc-1) run_params.append("&");
		run_params.append(argv[i]);
	}
	return cprOk;
}

CmdProcResult handler_listen(int argc, const char *argv[]) {
	int t, i, rslt;
	cfgprm cfg;
	jmp_buf markbuf;
	void *extcfg = NULL;
	DbgSite* asite = NULL;
	
	if (fsite != NULL) {
	    fprintf(rl_outstream, msg_dbg_already);
	    return cprFailed;
	}

	cfg.port = port;
	cfg.breakonload = 1;
	cfg.breakonfinish = 1;

	rslt = listener.startJIT(&cfg, extcfg);
	DBG_TRACE(("startJIT->%d\n", rslt));
	if (!rslt) {
	    fprintf(rl_outstream, "Failed to start listener.Error (%d)\n", listener.errorJIT());
	    return cprFailed;
	}
	if (setjmp(markbuf) != 0) {
	    listener.stopJIT();
	    return cprFailed;
	}
	fprintf(rl_outstream, "Waiting for debug session (press Ctrl-C to stop)...");
	fflush(rl_outstream);
	if (argc >= 1)
	    t = atoi(argv[0]);
	else 
	    t = -1;
	HRESULT hr = E_FAIL;
	for (i = 0; i != t; i++) {
	    if (async_pause) {
		async_pause = false;
		break;
	    }
    	    hr = listener.peekJIT(0, 999, 15000, NULL, SOF_SEND_OUTPUT|SOF_BREAKONLOAD|SOF_BREAKONFINISH, NULL, asite);
	    if (SUCCEEDED(hr)) 
		break;
	}
        if (FAILED(hr)) {
	    delete asite;
	}
	fprintf(rl_outstream, "\n");
	listener.stopJIT();
	return cprOk;
}

CmdProcResult handler_kill(int argc, const char *argv[]) {
	UIAction = BP_ABORT;
	return cprOk;
}

CmdProcResult handler_quit(int argc, const char *argv[]) {
	UIAction = BP_ABORT;
	return cprOkQuit;
}


// STEPPING

CmdProcResult handler_step(int argc, const char *argv[]) {
	UIAction = BP_STEP_INTO;
	return cprOk;
}

CmdProcResult handler_next(int argc, const char *argv[]) {
	UIAction = BP_STEP_OVER;
	return cprOk;
}

CmdProcResult handler_finish(int argc, const char *argv[]) {
	UIAction = BP_STEP_OUT;
	return cprOk;
}

CmdProcResult handler_cont(int argc, const char *argv[]) {
	UIAction = BP_CONTINUE;
	return cprOk;
}


// BACKTRACE

CmdProcResult print_bt_item(int argc, const char *argv[], int incdir) {
	HRESULT hr;
	SCOPELIST scopelist;
	SCOPELIST::iterator it;
	long v = 1;
	char *p;
	std::string mod_name, warn;

	if (fsite == NULL) return cprFailed;

	hr=fsite->requestor()->getscopelist(&scopelist);
	if (FAILED(hr) || scopelist.size() == 0) {
		fprintf(rl_outstream, "No stack\n");
		return cprOk;
	}

	if (argc==1) v = strtol (argv[0], &p, 0);
	if (incdir != 0) {
		if (incdir < 0) v = -v;
		bt_curr_idx += v;
	} else {
		if (argc > 0) 
			bt_curr_idx = v;
	}
	if (bt_curr_idx >= (int)scopelist.size()) bt_curr_idx = scopelist.size() - 1;
	if (bt_curr_idx < 0) bt_curr_idx = 0;
	it = scopelist.begin() + bt_curr_idx;
	bt_curr_scopeid = (*it).scope_id;
	if (FAILED(fsite->requestor()->requestdocfilename((*it).mod_no, &mod_name))) {
		mod_name = "?";
	}
	if (!mapper.mapfilename(mkRemoteToLocal, mod_name, mod_name, warn)) {
		fprintf(rl_outstream, "%s\n", warn.c_str());
	}
	fprintf(rl_outstream, "#%ld <%s> at %s:%ld\n", bt_curr_idx, (*it).scope_descr.c_str(), mod_name.c_str(), (*it).line_no);
	return cprOk;
}

CmdProcResult handler_up(int argc, const char *argv[]) {
	return print_bt_item(argc, argv, 1);
}

CmdProcResult handler_down(int argc, const char *argv[]) {
	return print_bt_item(argc, argv, -1);
}

CmdProcResult handler_where(int argc, const char *argv[]) {
	int cnt;
	long v, start, i;
	HRESULT hr;
	SCOPELIST scopelist;
	SCOPELIST::iterator it;
	char *p;
	std::string mod_name, warn;

	if (fsite == NULL) return cprFailed;

	hr=fsite->requestor()->getscopelist(&scopelist);
	if (FAILED(hr) || scopelist.size() == 0) {
		fprintf(rl_outstream, "No stack\n");
		return cprOk;
	}
	start = 0;
	cnt = scopelist.size();
	if (argc==1) {
		v = strtol (argv[0], &p, 0);
		if (v == 0) return cprOk;
		if (v > 0) {
			cnt = v;
		} else {
			start = cnt + v - 1;
			if (start < 0) start = 0;
		}
	}
	for (i = 0, it = scopelist.begin() + start; (it!=scopelist.end() && i<cnt); it++,i++) {
		if (FAILED(fsite->requestor()->requestdocfilename((*it).mod_no, &mod_name))) {
			mod_name = "?";
		}
		if (!mapper.mapfilename(mkRemoteToLocal, mod_name, mod_name, warn)) {
			fprintf(rl_outstream, "%s\n", warn.c_str());
		}
		fprintf(rl_outstream, "#%ld <%s> at %s:%ld\n", i+start, (*it).scope_descr.c_str(), mod_name.c_str(), (*it).line_no);
	}
	return cprOk;
}

CmdProcResult handler_frame(int argc, const char *argv[]) {
	return print_bt_item(argc, argv, 0);
}


// BREAKPOINTS

CmdProcResult setBreakAt(const char *argv, bool istemp) {
	std::string mod_name, s, rmod_name, warn;
	long i;
	dbgint line_no, bp_no;
	char *p;
	HRESULT hr;

	mod_name = argv;
	i = mod_name.find_last_of(':');
	if (i > 0) {
		s.assign(mod_name,i+1,mod_name.length()-i);
		mod_name.erase(i);
		line_no = strtol (s.c_str(), &p, 0);
		if (p[0] != '\0') {
			fprintf(rl_outstream, "Wrong line number specified.\n");
			return cprWrongArg;
		}

		if (!mapper.mapfilename(mkLocalToRemote, mod_name, rmod_name, warn)) {
			fprintf(rl_outstream, "%s\n", warn.c_str());
		}

		hr = fsite->requestor()->setbreakpoint(line_no, 0, rmod_name, "", istemp, 0, &bp_no);
		if (!SUCCEEDED(hr)) {
			fprintf(rl_outstream, "Internal error.\n");
			return cprFailed;
		}
		if (istemp) 
			fprintf(rl_outstream, "TBreakpoint %ld : file %s, line %ld.\n",bp_no, mod_name.c_str(), line_no);
		else
			fprintf(rl_outstream, "Breakpoint %ld : file %s, line %ld.\n",bp_no, mod_name.c_str(), line_no);
	} else {
		fprintf(rl_outstream, "file name and line number must be separated by colon.\n");
		return cprWrongArg;
	}
	return cprOk;
}

CmdProcResult handler_break(int argc, const char *argv[]) {
	if (argc == 1) {
		return setBreakAt(argv[0], false);
	}
	return cprWrongArg;
}

CmdProcResult handler_tbreak(int argc, const char *argv[]) {
	if (argc == 1) {
		return setBreakAt(argv[0], true);
	}
	return cprWrongArg;
}

CmdProcResult break_setstate(dbgint bp_no, BPSTATE bps) {
	dbgint line_no, mod_no, skiphits, hitcount;
	std::string mod_name, condition;
	bool istemp, isunderhit;
	BPSTATE state;
	HRESULT hr;

	hr = fsite->requestor()->getbreakpoint(bp_no, &line_no, &mod_no, mod_name, condition, &istemp, &skiphits, &hitcount, &state, &isunderhit);
	if (!SUCCEEDED(hr)) {
		fprintf(rl_outstream, "Wrong breakpoint number specified.\n");
		return cprWrongArg;
	}
	hr = fsite->requestor()->updatebreakpoint(bp_no, line_no, condition, skiphits, bps);
	if (!SUCCEEDED(hr)) {
		fprintf(rl_outstream, "Internal error.\n");
		return cprFailed;
	}
	return cprOk;
}

CmdProcResult handler_enable(int argc, const char *argv[]) {
	char *p;

	dbgint bp_no = strtol (argv[0], &p, 0);
	if (p[0] != '\0') {
		fprintf(rl_outstream, "Wrong breakpoint number specified.\n");
		return cprWrongArg;
	}
	return break_setstate(bp_no, BPS_ENABLED);
}

CmdProcResult handler_disable(int argc, const char *argv[]) {
	char *p;

	dbgint bp_no = strtol (argv[0], &p, 0);
	if (p[0] != '\0') {
		fprintf(rl_outstream, "Wrong breakpoint number specified.\n");
		return cprWrongArg;
	}
	return break_setstate(bp_no, BPS_DISABLED);
}

CmdProcResult handler_delete(int argc, const char *argv[]) {
	char *p;

	dbgint bp_no = strtol (argv[0], &p, 0);
	if (p[0] != '\0') {
		fprintf(rl_outstream, "Wrong breakpoint number specified.\n");
		return cprWrongArg;
	}
	return break_setstate(bp_no, BPS_DELETED);
}

CmdProcResult handler_condtion(int argc, const char *argv[]) {
	char *p;
	dbgint line_no, mod_no, skiphits, hitcount;
	std::string mod_name, condition;
	bool istemp, isunderhit;
	BPSTATE state;
	HRESULT hr;

	if (argc == 0) {
		fprintf(rl_outstream, "Breakpoint number expected.\n");
		return cprWrongArg;
	}
	dbgint bp_no = strtol (argv[0], &p, 0);
	if (p[0] != '\0') {
		fprintf(rl_outstream, "Wrong breakpoint number specified.\n");
		return cprWrongArg;
	}
	hr = fsite->requestor()->getbreakpoint(bp_no, &line_no, &mod_no, mod_name, condition, &istemp, &skiphits, &hitcount, &state, &isunderhit);
	if (!SUCCEEDED(hr)) {
		fprintf(rl_outstream, "Wrong breakpoint number specified.\n");
		return cprWrongArg;
	}
	condition = "";
	for (int i=1; i<argc; i++) {
		condition.append(argv[i]);
	}
	hr = fsite->requestor()->updatebreakpoint(bp_no, line_no, condition, skiphits, state);
	if (!SUCCEEDED(hr)) {
		fprintf(rl_outstream, "Internal error.\n");
		return cprFailed;
	}
	return cprOk;
}



// EVALUATION

void getplainevalstr(const peval_item it, const std::string &name, std::string *str, bool resolveref);

void getplainarraystr(const EVALLIST *elist, std::string *str) {
	std::string tmp;
	if (!elist) {
		*str="";
		return;
	}
	for (EVALLIST::const_iterator it=elist->begin(); it!=elist->end(); it++) {
		getplainevalstr(*it, (*it)->name, &tmp, false);		
		str->append(tmp);
		str->append(", ");
	}
	str->erase(str->length()-2, 2);
}

void getplainevalstr(const peval_item it, const std::string &name, std::string *str, bool resolveref) {
	std::string tmp;
	*str = "";
	switch (it->valtypeid) {
		case ptObject:
			std_sprintf(*str, "%s = (%s) ", name.c_str(), it->valtype.c_str());
			getplainarraystr(it->subitems, &tmp);
			str->append(tmp);
			break;
		case ptArray:
			getplainarraystr(it->subitems, &tmp);
			std_sprintf(tmp, "%s = {%s}", name.c_str(), tmp.c_str());
			str->append(tmp);
			break;
		case ptDouble:
		case ptLong:
		case ptString:
		case ptBool:
			std_sprintf(*str, "%s = %s", name.c_str(), it->value.c_str());
			break;
		case ptResource:
			std_sprintf(*str, "%s = (%s) %s", name.c_str(), it->valtype.c_str(), it->value.c_str());
			break;
		case ptRef:
		case ptSoftRef:
			if (resolveref) {
				getplainevalstr(it->ref, name, str, false);
			} else {
				std_sprintf(*str, "%s = {reference}", name.c_str());
			}
			break;
		default:
			std_sprintf(*str, "%s = NULL", name.c_str());
	}	
}

void dodisplaywatch(int idx, bool showerr) {
	if (sess_state != ssStarted || !fsite) {
		fprintf(rl_outstream, "No symbol table is loaded.  Use the \"file\" command.\n");
		return;
	}
	for (watch_list::iterator it=watches.begin(); it!=watches.end(); it++) {
		if (it->idx==idx || idx==-1) {
			HRESULT hr;
			std::string ser_str, err_str, str;
			hr = fsite->requestor()->evaluate(it->eval_str, bt_curr_scopeid, &ser_str, &err_str);
			if (hr != S_OK) {
				if (showerr)
					fprintf(rl_outstream, "%d: Invalid expression (%s)\n",it->idx,err_str.c_str());
			} else {
				EVALLIST elist(ser_str, it->eval_str);
				if (!elist.empty()) {
					getplainevalstr(elist[0], it->eval_str.c_str(), &str, true);
				} else {
					str = "Internal error";
				}
				fprintf(rl_outstream, "%d: %s\n",it->idx, str.c_str());
			}
			if (idx!=-1) break;
		}
	}
}

CmdProcResult handler_display(int argc, const char *argv[]) {
	if (sess_state != ssStarted) {
		if (argc != 0)
			fprintf(rl_outstream, "No symbol table is loaded.  Use the \"file\" command.\n");
		return cprOk;
	} else {
		if (argc==0) {
			dodisplaywatch(-1,true);
		} else {
			watch_rec rec;
			rec.idx = ++last_watch_idx;
			rec.eval_str = argv[0];
			watches.push_back(rec);
			dodisplaywatch(rec.idx,true);
		}
	}
	return cprOk;
}

CmdProcResult handler_undisplay(int argc, const char *argv[]) {		
	if (sess_state != ssStarted) {
		fprintf(rl_outstream, "No symbol table is loaded.  Use the \"file\" command.\n");
		return cprOk;
	} else {
		if (argc==0) {
			watches.clear();
		} else {
			for (int i = 0; i < argc; i++) {
				int idx = atoi(argv[i]);
				for (watch_list::iterator it=watches.begin(); it!=watches.end(); it++) {
					if (it->idx==idx) {
						watches.erase(it);
						break;
					}
				}
			}
		}
	}
	return cprOk;
}

CmdProcResult handler_print(int argc, const char *argv[]) {
	if (sess_state != ssStarted || !fsite) {
		fprintf(rl_outstream, "No symbol table is loaded.  Use the \"file\" command.\n");
		return cprOk;
	}
	HRESULT hr;
	std::string ser_str, err_str, str;
	str = argv[0];
	hr = fsite->requestor()->evaluate(str, bt_curr_scopeid, &ser_str, &err_str);
	if (hr != S_OK) {
		//if (showerr)
		fprintf(rl_outstream, "%ld: Invalid expression (%s)\n",++print_counter, err_str.c_str());
	} else {
		EVALLIST elist(ser_str, str);
		if (!elist.empty()) {
			getplainevalstr(elist[0], "", &str, true);
		} else {
			str = "Internal error";
		}
		fprintf(rl_outstream, "$%ld%s\n",++print_counter, str.c_str());
	}
	return cprOk;
}

CmdProcResult handler_output(int argc, const char *argv[]) {
	if (sess_state != ssStarted || !fsite) {
		fprintf(rl_outstream, "No symbol table is loaded.  Use the \"file\" command.\n");
		return cprOk;
	}
	std::string output;
	DBG_TRACE(("output(%s)\n", argc > 0 ? argv[0] : ""));
	fsite->get_output(output);
	print_lines(rl_outstream, output, 1, count_lines(output), 0);
	return cprOk;
}
// INFORMATION

CmdProcResult handler_help(int argc, const char *argv[]) {
	CommandEntry *cmd = NULL;
	if (argc > 0) {
		CommandList *lst = &main_cmdlist;
		int i = 0;
		while (i<argc && lst!=NULL) {
			cmd = lst->lookup_cmd(argv[i]);
			if (cmd == NULL) break;
			lst = cmd->subCommandList;
			i++;
		}
		if (cmd == NULL) {
			fprintf(rl_outstream, "Undefined sub-command: \"%s\".  Try \"help\".\n", argv[i]);
			return cprOkIgnored;
		} else {
			fprintf(rl_outstream, "%s\n", cmd->help);
		}
	} else {
		CommandList::iterator it;
		std::string s;
		for (it = main_cmdlist.begin(); it != main_cmdlist.end(); it++) {
			int cnt, l;
			s = it->name + std::string("\t");
			l = ((strlen(it->name) + 8) & ~7) + 1;
			cnt=(17 - l) / 8;
			if (cnt > 0) s.append(cnt, '\t');
			s.append(it->help);
			s.append("\n\n");
			fputs(s.c_str(), rl_outstream);
		}
	}
	return cprOk;
}

CmdProcResult handler_info_sources(int argc, const char *argv[]) {
	if (sess_state != ssStarted) {
		if (main_script.length() != 0) {
			fprintf(rl_outstream, "Source files for which symbols have been read in:\n\n%s\n",main_script.c_str());
			return cprOk;
		} else {
			fprintf(rl_outstream, "No symbol table is loaded.  Use the \"file\" command.\n");
			return cprFailed;
		}
	} else {
		DOCLIST doclist;
		DOCLIST::iterator it;
		std::string mod_name, warn;
		if (fsite == NULL || FAILED(fsite->requestor()->getdoclist(&doclist))) return cprFailed;
		for (it=doclist.begin(); it!=doclist.end(); it++) {
			if (it!=doclist.begin()) fprintf(rl_outstream, ",\n");
			if (!mapper.mapfilename(mkRemoteToLocal, (*it).mod_name, mod_name, warn)) {
				fprintf(rl_outstream, "%s\n",warn.c_str());				
			}
			fputs((*it).mod_name.c_str(), rl_outstream);
		}
		fprintf(rl_outstream, "\n");
	}
	return cprFailed;
}

CmdProcResult handler_info_source(int argc, const char *argv[]) {
	if (sess_state != ssStarted) {
		fprintf(rl_outstream, "No current source file.\n");
		return cprOk;
	} else {
		std::string mod_name, scope_name, warn;
		dbgint mod_no, line_no;
		
		if (fsite == NULL || 
			FAILED(fsite->requestor()->getcurloc(&mod_no, &line_no, &scope_name)) ||
			FAILED(fsite->requestor()->requestdocfilename(mod_no, &mod_name))) return cprFailed;
		if (!mapper.mapfilename(mkRemoteToLocal, mod_name, mod_name, warn)) {
			fprintf(rl_outstream, "%s\n",warn.c_str());				
		}
		fprintf(rl_outstream, "Current source file is %s\nLocated in %s\n", mod_name.c_str(), mod_name.c_str());
	}
	return cprFailed;
}

CmdProcResult handler_info_program(int argc, const char *argv[]) {
	if (sess_state != ssStarted) {
		fprintf(rl_outstream, "The program being debugged is not being run.\n");
		return cprOk;
	} else {
	}
	return cprOk;
}

void print_breakpoint(dbgint req_bp_no, bool *hdrprinted) {
	BREAKPOINTLIST bplist;
	const char *disp, *enb;
	bool anyprinted = false;
	std::string mod_name, warn;

	fsite->requestor()->getbreakpointlist(&bplist);
	for (BREAKPOINTLIST::iterator it = bplist.begin(); it!=bplist.end(); it++) {
		if (req_bp_no < 0 || req_bp_no == it->bp_no) {
			if (!anyprinted) {
				anyprinted = true;
				if (!*hdrprinted) {
					fprintf(rl_outstream, "Num  Type       Disp  Enb What\n");
					*hdrprinted = true;
				}
			}
			disp = (it->istemp || (it->state &~BPS_UNRESOLVED) == BPS_DELETED) ? ("del") : ("keep");
			enb = ((it->state &~BPS_UNRESOLVED) == BPS_DISABLED) ? ("n") : ("y");
			if (!mapper.mapfilename(mkRemoteToLocal, it->mod_name, mod_name, warn)) {
				fprintf(rl_outstream, "%s\n",warn.c_str());				
			}			
			fprintf(rl_outstream, "%-4d breakpoint %s  %s   at %s:%ld\n", it->bp_no, disp, enb, mod_name.c_str(), it->line_no);
			if (it->condition.length()) {
				fprintf(rl_outstream, "          stop only if %s\n", it->condition.c_str());
			}
			if (it->skiphits > 0) {
				fprintf(rl_outstream, "          ignore next %d hits\n", it->skiphits - it->hitcount);
			}
		}
	}
	if (!anyprinted ) {
		if (req_bp_no>=0)
			fprintf(rl_outstream, "Invalid breakpoint number %ld\n",req_bp_no);
		else
			fprintf(rl_outstream, "No breakpoints or watchpoints.\n");
	}
}

CmdProcResult handler_info_breakpoints(int argc, const char *argv[]) {
	char *p;
	bool hdrprinted = false;

	if (argc == 0) 
		print_breakpoint(-1, &hdrprinted);
	else {
		for (int i=0; i<argc; i++) {
			dbgint bp_no = strtol (argv[i], &p, 0);
			print_breakpoint(bp_no, &hdrprinted);
		}
	}
	return cprOk;
}

CmdProcResult handler_pwd(int argc, const char *argv[]) {
	char dir[PATH_MAX];
	getcwd(dir,sizeof(dir));
	fprintf(rl_outstream, "Working directory is %s.\n", dir);
	return cprOk;
}

CmdProcResult handler_list(int argc, const char *argv[]) {

	if (argc == 1 && argv[0] && stricmp(argv[0], "auto") == 0) {
	    dbg_AUTO_LIST = !dbg_AUTO_LIST;
	    fprintf(rl_outstream, "list auto is %s\n", dbg_AUTO_LIST ? "on" : "off");
    	    return cprOk;
	}
	
	if (fsite == NULL || fsite->curr_line_no() == 0 || fsite->curr_mod_no() == 0) {
		fprintf(rl_outstream, "No symbol table is loaded.  Use the \"file\" command.\n");
		return cprFailed;
	} else {
		std::string src;
		int startline, cnt;
		fsite->get_curr_src(src);
		startline = fsite->curr_line_no() - 2;
		cnt = 5;
		if (startline < 1) {
		    cnt += startline - 1;
		    startline = 1;
		}
		print_lines(rl_outstream, src, startline, cnt, fsite->curr_line_no());
	}
	return cprOk;
}


//////////////              //////////////
//               READLINE 			    //
//////////////______________//////////////


/*bool input_avail(FILE *instream) {
#ifdef WIN32
	int rv;
	HANDLE fno;
	fno = GetStdHandle(STD_INPUT_HANDLE);
	rv = WaitForSingleObjectEx(fno, 500, TRUE);
	return (rv == WAIT_OBJECT_0);
#else
    fd_set readfds, exfds;
    struct timeval timeout;
    int fno, rv;
    fno = fileno(instream);
    FD_ZERO(&readfds);
    FD_ZERO(&exfds);
    FD_SET(fno, &readfds);
    FD_SET(fno, &exfds);
    timeout.tv_sec = 0;
    timeout.tv_usec = 50000;
    rv = select(fno+1, &readfds, NULL, &exfds, &timeout);
    return (rv>0);
#endif
}*/

void read_cfg(const std::string &fname) {
	FILE *fp;
	char *p, *t;
	char buf[1024];
	size_t len;
	
	if (!fileexists(fname))
	    return;
	fp = fopen(fname.c_str(), "r");
	if (fp == NULL)
	    return;
	while (fgets(buf, sizeof(buf), fp)) {
	    if (buf[0] == '#')
		continue;
	    p = buf;
	    while (*p == ' '|| *p == '\t') p++;
	    if (p[0] == '#' || p[0] == ';')
		continue;
	    len = strlen(p);
	    t = p + len - 1;
	    while (t > p && (*t == ' ' || *t == '\t' || *t == '\n')) t--;
	    t[1] = '\000';
	    len = t - p;
	    if (len == 0)
		continue;

	    DBG_TRACE(("config option->\"%s\"\n", p));
	    process_line(p);
	}
	fclose(fp);
}


void process_loop(unsigned short port, char *execarg, char *cgihandler, int mode,
 char *mapurlroot, char *maplocalroot, char *mapremoteroot) {
	jmp_buf markbuf;
	char *aline;
	std::string hist_fname, cfg_fname;
	
	hist_fname.assign(getenv("HOME"));
	cfg_fname.assign(hist_fname);
	hist_fname.append("/.dbgcli");
	cfg_fname.append("/.dbgcli_init");
	
	rl_initialize();
	read_history((char*)hist_fname.c_str());
//	log = fopen(logfile,"w");
	
	signal(SIGTERM, handle_sigterm);
#if defined(SIGIO)
	signal(SIGIO, handle_sigio);
#endif
	signal(SIGINT, handle_sigint);
#if defined(SIGPIPE) && defined(SIG_IGN)
	signal(SIGPIPE, SIG_IGN);
#endif

	if (setjmp(markbuf) != 0) {
		return;
	}
	read_cfg(cfg_fname);
//	rl_initialize();
//	rl_set_signals();
//	rl_callback_handler_install(DBG_PROMPT, (void(*)())process_line);
	if (mode == 1 || mode == 0) {
		SimpleStringArray ssa;
		ssa.add((mode == 1) ? "on" : "off");
		handler_set_mode(ssa.count(), ssa.array());
	}
	if (execarg) {
		SimpleStringArray ssa;
		ssa.add(execarg);
		handler_file(ssa.count(), ssa.array());
	}
	if (cgihandler) {
		SimpleStringArray ssa;
		ssa.add(cgihandler);
		handler_set_cgiexec(ssa.count(), ssa.array());
	}
	if (port > 0) {
		SimpleStringArray ssa;
		char numbuf[20];
		snprintf(numbuf, sizeof(numbuf), "%d", port);
		ssa.add(numbuf);
		handler_set_port(ssa.count(), ssa.array());
	}
	if (mapurlroot) {
		SimpleStringArray ssa;
		ssa.add(mapurlroot);
		handler_set_mapurlroot(ssa.count(), ssa.array());
	}
	if (maplocalroot) {
		SimpleStringArray ssa;
		ssa.add(maplocalroot);
		handler_set_maplocalroot(ssa.count(), ssa.array());
	}
	if (mapremoteroot) {
		SimpleStringArray ssa;
		ssa.add(mapremoteroot);
		handler_set_mapremoteroot(ssa.count(), ssa.array());
	}

	while (!quit_request) {
		if (run_request) {
			if (listener.numberofclients() > 0) {
				listener.stop(true);
			}
			run_request = false;
			do_run();
		}
		if (listener.numberofclients() > 0) {
			for (int cl=0; cl < listener.numberofclients(); cl++) {
				ThreadBase *th = listener.client(cl);
				DBG_TRACE(("calling ThreadBase[%d]::process()>\n", cl));
				th->process();
				DBG_TRACE(("calling ThreadBase[%d]::process()<\n", cl));
			}
			DBG_TRACE(("calling listener(0x%0.8x).peekJIT()<\n", (int)&listener));
			HRESULT hr = listener.peekJIT(0, 1, 1, NULL, SOF_SEND_OUTPUT|SOF_BREAKONLOAD|SOF_BREAKONFINISH);
			if (SUCCEEDED(hr)) {
				DBG_TRACE(("[dbg] a JIT client found\n"));
			} else {
				listener.cleanup();
			}

		} else {
			MEMCHK_OFF;
			aline = readline(DBG_PROMPT);
			MEMCHK_ON;			
			process_line(aline);
			free(aline);
//			if (input_avail(rl_instream)) {
//				rl_callback_read_char();
//			}
		}
	}
//	rl_callback_handler_remove();
//	rl_clear_signals();
	write_history((char*)hist_fname.c_str());
	if (log) fclose(log);
}


/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 */
