/***************************************************************************
                          command.h  -  description
                             -------------------
    begin                : Sun Aug 5 2001
    author               : Dmitri Dmitrienko
    copyright            : (C) 2001, 2007, Dmitri Dmitrienko
    www                  : http://dd.cron.ru
    license              : This source file is subject to version 1.01 of 
                           the License,  that is bundled with this package 
                           in the file LICENSE, and is available at through 
                           the world-wide-web at http://dd.cron.ru/license
 ***************************************************************************/

#ifndef _COMMAND_H_
#define _COMMAND_H_

#ifdef WIN32
#include <config.w32.h>
#else
#include <config.h>
#endif

#include <algorithm>
#include <vector>
#include <string.h>

typedef enum tagCmdProcResult {
	cprOk			= 0x0000,
	cprOkQuit		= 0x0001,
	cprOkEmpty		= 0x0002,
	cprOkIgnored	= 0x0003,
	cprOkExtraArg	= 0x0004,
// negative
	cprUnknownCmd	= -0x0001,
	cprArgExpexted	= -0x0002,
	cprOutRangeVal	= -0x0003,
	cprNoArg		= -0x0004,
	cprWrongArg		= -0x0005,
	cprFailed		= -0x0006,
} CmdProcResult;
 
typedef enum tagHandlerType {
	htAnyArgs				= 0x0000,
	htNoArg					= 0x0001,
	htOneOrNoneStrArg		= 0x0002,
	htOneOrNoneUShortArg	= 0x0003,
	htOneStrArg				= 0x0004,
	htOneUShortArg			= 0x0005,
	htOneOrNoneShortArg		= 0x0006,
	htNoneOrManyUShortArg	= 0x0007,
} HandlerType;

typedef enum tagSessionState {
	ssNotStarted	= 0x0000,
	ssStarted		= 0x0001,
	ssStopped		= 0x0002,
} SessionState;

typedef CmdProcResult (*cmdFunc)(int argc, const char *argv[]);

class CommandList;

typedef struct tagCommandEntry {
	char *name;
	bool reqSessStarted;
	HandlerType ht;
	CommandList* subCommandList;
	cmdFunc func;
	char *help;
} CommandEntry;

#define ENDCOMMAND {NULL, 0, htNoArg, NULL, NULL}

typedef std::vector<CommandEntry> CommandListBase;

class CommandList: public CommandListBase {
public:
	CommandList(CommandEntry *list) {
		for (CommandEntry *entry = list; entry && entry->name; entry++) {
			iterator it;
			it = (iterator)(std::lower_bound(begin(), end(), *entry));
			insert(it, 1, *entry);
		}
	}
	CommandEntry *lookup_cmd(const char *cmd);
};

#define safestricmp(n1, n2)	\
	((n1 == n2) ? 0 : \
	((!n1) ? (-1) : \
	((!n2) ? (1) : \
	(strcasecmp(n1,n2)))))

inline bool operator < (const CommandEntry& e1, const CommandEntry& e2) {
	return (safestricmp(e1.name, e2.name) < 0);
}

inline bool operator == (const CommandEntry& e1, const CommandEntry& e2) {
    return (safestricmp(e1.name, e2.name) == 0);
}

inline bool operator <= (const CommandEntry& e1, const CommandEntry& e2) {
    return (safestricmp(e1.name, e2.name) <= 0);
}


// SETTINGS
CmdProcResult handler_set_port(int argc, const char *argv[]);
CmdProcResult handler_show_port(int argc, const char *argv[]);
CmdProcResult handler_set_mode(int argc, const char *argv[]);
CmdProcResult handler_show_mode(int argc, const char *argv[]);
CmdProcResult handler_set_cgiexec(int argc, const char *argv[]);
CmdProcResult handler_show_cgiexec(int argc, const char *argv[]);
CmdProcResult handler_set_mapurlroot(int argc, const char *argv[]);
CmdProcResult handler_show_mapurlroot(int argc, const char *argv[]);
CmdProcResult handler_set_maplocalroot(int argc, const char *argv[]);
CmdProcResult handler_show_maplocalroot(int argc, const char *argv[]);
CmdProcResult handler_set_mapremoteroot(int argc, const char *argv[]);
CmdProcResult handler_show_mapremoteroot(int argc, const char *argv[]);
// SESSION
CmdProcResult handler_file(int argc, const char *argv[]);
CmdProcResult handler_run(int argc, const char *argv[]);
CmdProcResult handler_listen(int argc, const char *argv[]);
CmdProcResult handler_kill(int argc, const char *argv[]);
CmdProcResult handler_quit(int argc, const char *argv[]);
// STEPPING
CmdProcResult handler_step(int argc, const char *argv[]);
CmdProcResult handler_next(int argc, const char *argv[]);
CmdProcResult handler_finish(int argc, const char *argv[]);
CmdProcResult handler_cont(int argc, const char *argv[]);
// BACKTRACE
CmdProcResult handler_up(int argc, const char *argv[]);
CmdProcResult handler_down(int argc, const char *argv[]);
CmdProcResult handler_where(int argc, const char *argv[]);
CmdProcResult handler_frame(int argc, const char *argv[]);
// BREAKPOINTS
CmdProcResult handler_break(int argc, const char *argv[]);
CmdProcResult handler_tbreak(int argc, const char *argv[]);
CmdProcResult handler_enable(int argc, const char *argv[]);
CmdProcResult handler_disable(int argc, const char *argv[]);
CmdProcResult handler_delete(int argc, const char *argv[]);
CmdProcResult handler_condtion(int argc, const char *argv[]);
// EVALUATION
CmdProcResult handler_display(int argc, const char *argv[]);
CmdProcResult handler_undisplay(int argc, const char *argv[]);
CmdProcResult handler_print(int argc, const char *argv[]);
CmdProcResult handler_output(int argc, const char *argv[]);
// INFORMATION
CmdProcResult handler_help(int argc, const char *argv[]);
CmdProcResult handler_info_sources(int argc, const char *argv[]);
CmdProcResult handler_info_source(int argc, const char *argv[]);
CmdProcResult handler_info_program(int argc, const char *argv[]);
CmdProcResult handler_info_breakpoints(int argc, const char *argv[]);
CmdProcResult handler_pwd(int argc, const char *argv[]);
CmdProcResult handler_list(int argc, const char *argv[]);



//
void process_loop(unsigned short port, char *execarg, char *cgihandler, int mode,
    char *mapurlroot, char *maplocalroot, char *mapremoteroot);

#endif
