/***************************************************************************
                          main.cpp  -  description
                             -------------------
    begin                : Sun Aug 5 2001
    author               : Dmitri Dmitrienko
    copyright            : (C) 2001, 2007, Dmitri Dmitrienko
    www                  : http://dd.cron.ru
    license              : This source file is subject to version 1.01 of 
                           the License,  that is bundled with this package 
                           in the file LICENSE, and is available at through 
                           the world-wide-web at http://dd.cron.ru/license
 ***************************************************************************/

#if defined(WIN32) && defined(_DEBUG)
#define _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#include <crtdbg.h>
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#ifdef WIN32
#include <io.h>
#endif

#include <getopt.h>

#include <signal.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include "dbgbase/dbgbase_globals.h"
#include "dbgbase/dbg_net.h"

#include "command.h"

#ifdef WIN32
#include "config.w32.h"
#endif


#ifndef DBG_CONFIGFILE
#define DBG_CONFIGFILE "/etc/dbg-cli.ini"
#endif

static char *execarg = NULL;
static char *cgihandler = NULL;
static char *selfname = NULL;
static char *mapurlroot = NULL;
static char *maplocalroot = NULL;
static char *mapremoteroot = NULL;
static unsigned short port = 0;
static int mode = -1;
static int print_help = 0;
static int print_version = 0;


static const char short_options[] = "hvp:c:m:U:R:L:f:";
static struct option long_options[] = {
    {"help", no_argument, 0, 'h'},
    {"version", no_argument, 0, 'v'},
    {"port", required_argument, 0, 'p'},
    {"cgi", required_argument, 0, 'c'},
    {"mode", required_argument, 0, 'm'},
    {"mapurlroot", required_argument, 0, 'U'},
    {"mapremoteroot", required_argument, 0, 'R'},
    {"maplocalroot", required_argument, 0, 'L'},
    {"file", required_argument, 0, 'f'},
    {0, no_argument, 0, 0}
};


#define dbg_cmdline_help "\
Usage:\n\n\
    %s [options] [URL or script-file]\n\n\
Options:\n\n\
    -p, --port=PORT		Use PORT value for listening\n\
    -c, --cgi=PHPEXEC		Set PHPEXEC to act as php \n\
				executable for CGI mode\n\
    -m, --mode=N		Set run mode. N should be 0 for\n\
				CGI mode and 1 for HTTP one\n\
    -L, --maplocalroot=PATH	Set PATH as local root for mapping in HTTP mode\n\
    -R, --mapremoteroot=PATH	Set PATH as remote root for mapping in HTTP mode\n\
    -U, --mapurlroot=URL	Set URL as root url for mapping in HTTP mode\n\
    -h, --help			Display this help and exit\n\
    -v, --version		Display version info and exit\n\
\n"

void print_dbg_help() {
	printf(dbg_cmdline_help, selfname);
}

int parse_args(int ac, char * const av[], int *option_index, int is_cfg_file) {
    char c;
    
    DBG_TRACE(("parse_args(argc=%d, argv=%s)\n", ac, SON(av[1])));
    if (ac <= 1)
	return 0;
	
    optind = 0;
    opterr = !is_cfg_file; /* warn only if not cfg_file */
    
    while (1) {
	c = getopt_long (ac, av, short_options, long_options, option_index);
        DBG_TRACE(("arg#=%d, c=%d (%c)\n", *option_index, c, c));
	if (c == EOF) 
	    break;

	if (c == 0 && long_options[*option_index].flag == 0) {
	    c = long_options[*option_index].val;
	}
		
	switch (c) {
	    case 0:
		continue;
	    case 'h':
		if (!is_cfg_file)
		    print_help = 1;
		break;
	    case 'v':
		if (!is_cfg_file)
		    print_version = 1;
		break;
	    case 10:
		DBG_TRACE(("execarg->%s\n", optarg));
		if (execarg != NULL) 
		    free(execarg);
		execarg = strdup(optarg);
		break;
	    case 'p': {
		unsigned long i;
		char *p;
				
		DBG_TRACE(("port->%s\n", optarg));
		i = strtoul (optarg, &p, 0);
		if ((i == 0 && p == optarg) || i == 0 || i > 0xFFFF)
		    fprintf(stderr,"warning: could not set 'port' to `%s'.\n", optarg);
		else
		    port = (unsigned short)i;
		}
		break;
	    case 'c':
		DBG_TRACE(("cgi handler->%s\n", optarg));
		if (!optarg || !optarg[0])
		    fprintf(stderr,"warning: could not set CGI handler to be empty.\n");
		else {
		    if (cgihandler != NULL)
			free(cgihandler);
		    cgihandler = strdup(optarg);
		}
		break;
	    case 'U':
		DBG_TRACE(("mapurlroot->%s\n", optarg));
		if (!optarg || !optarg[0])
		    fprintf(stderr,"warning: could not set mapurlroot to be empty.\n");
		else {
		    if (mapurlroot != NULL)
			free(mapurlroot);
		    mapurlroot = strdup(optarg);
		}
		break;
	    case 'L':
		DBG_TRACE(("maplocalroot->%s\n", optarg));
		if (!optarg || !optarg[0])
		    fprintf(stderr,"warning: could not set maplocalroot to be empty.\n");
		else {
		    if (maplocalroot != NULL)
			free(maplocalroot);
		    maplocalroot = strdup(optarg);
		}
		break;
	    case 'R':
		DBG_TRACE(("mapremoteroot->%s\n", optarg));
		if (!optarg || !optarg[0])
		    fprintf(stderr,"warning: could not set mapremoteroot to be empty.\n");
		else {
		    if (mapremoteroot != NULL)
			free(mapremoteroot);
		    mapremoteroot = strdup(optarg);
		}
		break;
	    case 'f':
		DBG_TRACE(("file->%s\n", optarg));
		if (!optarg || !optarg[0])
		    fprintf(stderr,"warning: could not set file to be empty.\n");
		else {
		    if (execarg != NULL)
			free(execarg);
		    execarg = strdup(optarg);
		}
		break;
	    case 'm': {
		long i;
		char *p;
				
		DBG_TRACE(("mode->%s\n", optarg));
		i = strtol (optarg, &p, 0);
		if (i != 0 && i != 1)
		    fprintf(stderr,"warning: invalid mode specified `%s'.\n", optarg);
		else
		    mode = i;
		}
		break;
	    case '?':
		DBG_TRACE(("unknown arg->%s\n", optarg));
		if (!is_cfg_file) {
    		    fprintf(stderr, "Use `%s --help' for a complete list of options.\n", selfname);
		    exit (1);
		}
	    default:
		DBG_TRACE(("unhandled arg->%s\n", optarg));
	}
    }
    return 0;
}

int main(int argc, char *argv[]) {
	int count, oi = 0;
	jmp_buf markbuf;

#if defined(WIN32) && defined(_DEBUG)
	_CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);
#endif


#if defined(SIGPIPE) && defined(SIG_IGN)
	signal(SIGPIPE,SIG_IGN); 
#endif
	if (setjmp(markbuf) != 0) {
		return -1;
	}
	
	{
	    char *p = strrchr(argv[0], '/');
	    if (p == NULL)
		p = strrchr(argv[0], '\\');
	    if (p == NULL)
		p = argv[0];
	    else
		p++;
	    selfname = strdup(p);
	}
	
	printf(DBG_API_DESCRIPTION "\n");
	
//	readcfg();
	
	parse_args(argc, argv, &oi, 0);

	/* OK, that's all the options.  The other arguments are filenames.  */
	count = 0;
	for (; optind < argc; optind++) {
		switch (++count) {
			case 1:
				if (execarg != NULL)
				    free(execarg);
				execarg = strdup(argv[optind]);
				break;
			case 2:
				fprintf(stderr, "Extra command line argument(s) ignored. (%s%s)\n", 
					argv[optind], (optind == argc - 1) ? "" : " ...");
				break;
		}
	}

	if (print_version) {
		return 0;
	}
	if (print_help) {
		print_dbg_help();
		return 0;
	}

#if WIN32
	{
		WSADATA wsdata;
		WSAStartup(2,&wsdata);
	}
#endif

#ifdef WIN32
	_fmode = _O_BINARY;
	setmode(_fileno(stdin), O_BINARY);
	setmode(_fileno(stdout), O_BINARY);
	setmode(_fileno(stderr), O_BINARY);
#endif
	process_loop(port, execarg, cgihandler, mode, mapurlroot, maplocalroot, mapremoteroot);

#if WIN32
	WSACleanup();
#endif
	if (execarg != NULL)
	    free(execarg);
	if (cgihandler != NULL)
	    free(cgihandler);
	if (selfname != NULL)
	    free(selfname);
	if (mapurlroot != NULL)
	    free(mapurlroot);
	if (maplocalroot != NULL)
	    free(maplocalroot);
	if (mapremoteroot != NULL)
	    free(mapremoteroot);
	return 0;
}

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 */
