/***************************************************************************
                          DbgSess.cpp  -  description
                             -------------------
    begin                : Fri Mar 2 2001
    copyright            : (C) 2001 by Dmitri Dmitrienko
                         : (C) 2002, 2007 NuSphere Corp.
    www                  : http://dd.cron.ru
                         : http://www.nusphere.com/
    license              : This source file is subject to version 1.01 of 
                           the License,  that is bundled with this package 
                           in the file LICENSE, and is available at through 
                           the world-wide-web at http://dd.cron.ru/license
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef WIN32
#include "config.w32.h"
#include <windows.h>
#include <winsock2.h>
#include <direct.h>
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include "dbgbase/dbgbase_globals.h"
#include "dbgbase/dbgbase_classes.h"
#include "dbgbase/dbg_net.h"

#include "UrlParse.h"
#include "StrTable.h"
#include "DbgSpawn.h"


void GenerateSessionID(std::string& sessid) {
	char buf[256];
	struct timeval tv;
	static dbgint seq_id = 1;
	GETTIMEOFDAY(&tv);
	sprintf(buf, "%06ld%06ld%04d", tv.tv_sec, tv.tv_usec, seq_id++);
	sessid = buf;
}

void get_local_ip_list(std::string& ip) {
	struct hostent *he;
	char **addrlist = NULL;
	struct in_addr addr;
	char buf[256];


	if (gethostname(buf, sizeof(buf))!=0) strcpy(buf,"localhost");
	he = gethostbyname(buf);
	if (he != NULL) addrlist = he->h_addr_list;
	if (addrlist == NULL || *addrlist == NULL) {
		ip="127.0.0.1";
		return;
	}
	memset(&addr, 0, sizeof(addr));
	for (;*addrlist;addrlist++) {
		memcpy(&addr, *addrlist, he->h_length);
		ip = ip + inet_ntoa(addr) + ',';
	}
	ip.erase(ip.length()-1,1);
}

void http2cgi(const std::string& key, std::string& cgikey) {
	std::string::iterator it;
	cgikey = "HTTP_";
	cgikey.append(key);
	for (it=cgikey.begin(); it!=cgikey.end(); it++) {
		*it = toupper(*it);
		if (!isalnum(*it))  *it = '_';
	}
}

int add_common_cgi_vars(const SessRequest& req, StringTable& cgienv) {
	StringTableBase::const_iterator it;
	std::string cgikey, fpath, qry;
	int pos;
	char *p;

	fpath = req.request;
	pos = fpath.find("?");
	if (pos >= 0) {
		qry = fpath.substr(pos+1, fpath.length());
		fpath.erase(pos, fpath.length());
	}

	if (!findfile(fpath))
		return -1;

	cgienv.clear();
	cgienv.merge("PATH", getenv("PATH"));
	if ((p = getenv("PHPRC"))) cgienv.merge("PHPRC", p);

#ifdef WIN32
	if (p = getenv("COMSPEC")) cgienv.merge("COMSPEC", p);
	if (p = getenv("SystemRoot")) cgienv.merge("SystemRoot", p);
	if (p = getenv("WINDIR")) cgienv.merge("WINDIR", p);
#else
	if ((p = getenv("LD_LIBRARY_PATH"))) cgienv.merge("LD_LIBRARY_PATH", p);
#endif

	cgienv.merge("REDIRECT_QUERY_STRING", qry);
	cgienv.merge("REDIRECT_STATUS", "200");
	cgienv.merge("REDIRECT_URL", fpath);

	cgienv.merge("SERVER_SOFTWARE", "DBG/2.1" );
	cgienv.merge("SERVER_NAME", "localhost");
	cgienv.merge("SERVER_ADDR", "127.0.0.1");
	cgienv.merge("SERVER_PORT", "80");
	cgienv.merge("REMOTE_ADDR", "127.0.0.1");
	cgienv.merge("SCRIPT_FILENAME", req.cgi_handler);
	cgienv.merge("GATEWAY_INTERFACE", "CGI/1.1");
	cgienv.merge("SERVER_PROTOCOL", "HTTP/1.1");
	cgienv.merge("REQUEST_METHOD", "GET");
	cgienv.merge("QUERY_STRING", qry);
	cgienv.merge("REQUEST_URI", req.request);
	cgienv.merge("PATH_INFO", fpath);
	cgienv.merge("PATH_TRANSLATED", fpath);

	for (it = req.headers.begin(); it != req.headers.end(); it++) {
		http2cgi(it->key, cgikey);
		cgienv.merge(cgikey, it->val);
	}


	return 0;
}

HRESULT lookup_host(const char *addr, struct in_addr *in) {
	struct hostent *host_info;
	
#ifdef WIN32
	(*in).s_addr = inet_addr(addr);
	if ((*in).s_addr == INADDR_NONE) {
#else
	if (!inet_aton(addr, in)) {
#endif
		host_info = gethostbyname(addr);
		if (!host_info) {
			return E_FAIL;
		}
		*in = *((struct in_addr *) host_info->h_addr);
	}
	return S_OK;
}


HRESULT open_http_session(const SessRequest& req, SessConnection& conn, std::string& id) {
	UrlParse up = req.request;
	int sock = 0, rv;
	bool ok;
	struct sockaddr_in address;
	unsigned short port;

	DBG_TRACE(("open_http_session\n"));
	conn.result = DBGSE_WRONGREQ;
	if (!up.is_ok()) return E_FAIL;

	sock = socket(AF_INET, SOCK_STREAM, 0);
	conn.result = DBGSE_FAILEDCREATESOCKET;
	if (sock <= 0) return E_FAIL;

	memset(&address, 0, sizeof(address));
	address.sin_family = AF_INET;
	port = up.port();
	if (port==0) port = 80;
	address.sin_port = htons(port);

	conn.result = DBGSE_HTTPHOSTNOTFOUND;
	ok = SUCCEEDED(lookup_host(up.host().c_str(), &address.sin_addr));
	
	if (ok) {
		conn.result = DBGSE_HTTPHOSTNOTFOUND;
		while ((rv = connect(sock, (struct sockaddr *) &address,
			sizeof(address))) == SOCKET_ERROR && errno == EAGAIN);
		ok = (rv == 0);
	}

	if (ok) {
		StringTable hdrs;
		std::string str;

		hdrs.merge("", "GET " + up.path() + ((up.query().length() > 0) ? "?" + up.query() : "") + " HTTP/1.1");
		hdrs.merge("Accept", "*/*");
		hdrs.merge("Host", up.host());
		hdrs.merge("Connection", "Keep-Alive");		
		hdrs.merge(req.headers);
		hdrs.merge("Cookie", id, MERGE_APPEND_W_SEMICOLON);

		hdrs.getlines(str, ": ", "\r\n");
		str.append("\r\n");
		
		rv = send(sock, str.c_str(), str.length(), 0);
		ok = (rv == (int)str.length());
	}

	if (ok) {
		ok = (conn.receiver.assignsocket(sock) == 0);
	}

// finally OK
	if (ok) {
		return S_OK;
	}

// finally FAILED
	if (sock) {
		shutdown(sock, SD_BOTH);
		SCLOSE(sock);
	}
	conn.receiver.close();
	return E_FAIL;
}

HRESULT open_cgi_session(const SessRequest& req, SessConnection& conn, std::string& id) {
	StringTable cgienvtb;
	int ret_val = 0;
	pid_t cgi;
	pipe_t houtputpipe = INVALID_PIPE, herrpipe = INVALID_PIPE;
	HRESULT hr;

	DBG_TRACE(("open_cgi_session\n"));
	ret_val = add_common_cgi_vars(req, cgienvtb);
	if (ret_val != 0) {
		conn.result = DBGSE_FILENOTFOUND;
		return E_FAIL;
	}

	cgienvtb.merge("HTTP_COOKIE", id, MERGE_APPEND_W_SEMICOLON);

	SimpleStringArray env(cgienvtb);
	
	std::string hcgi = req.cgi_handler;
	if (!findfile(hcgi)) {
		conn.result = DBGSE_CGIHANDLERNOTFOUND;
		return  E_FAIL;
	}
	
	std::string path;
	char olddir[PATH_MAX];
	
	getcwd(olddir, sizeof(olddir));
	DBG_TRACE(("olddir=%s\n", olddir));
	
	extractpath(hcgi, path);
	DBG_TRACE(("path=%s\n", path.c_str()));
	
	chdir(path.c_str());
	
	
	cgi = spawn_process(hcgi.c_str(), (char**)env.array(), NULL, &houtputpipe, &herrpipe);
	
	chdir(olddir);
	
	hr = conn.assign_cgi(cgi, houtputpipe, herrpipe);
	DBG_TRACE(("open_cgi_session -> pid=%d, %d\n", cgi, hr));
	return hr;
}

HRESULT open_dbg_session(const SessRequest& req, SessConnection& conn) {
	std::string id, ip;
	char numbuf[16];
	unsigned short port;
	SERVER_TYPE server_type;

	DBG_TRACE(("open_dbg_session\n"));
	conn.result = DBGSE_SUCCESSFULL;
	GenerateSessionID(conn.sessid);
//	get_local_ip_list(ip);
	port = (req.port != 0) ? (req.port) : (DEFAULT_REQ_PORT);
	snprintf(numbuf, sizeof(numbuf), "%d", port);
//	id = DBGSESSVAR "=" + conn.sessid + "@" + ip + ":" + numbuf;
	id = DBGSESSVAR "=" + conn.sessid + "@clienthost:" + numbuf;
	server_type = req.server_type;
	if (server_type == SVR_AUTO) {
		std::string sch;
		UrlParse up(req.request);
		std_toupper(up.scheme(), sch);
		if (sch == "HTTP") server_type = SVR_HTTP;
			else if (sch == "HTTPS") server_type = SVR_HTTPS;
			else server_type = SVR_CGI;
	}
	switch (server_type) {
		case SVR_CGI: {
// CGI SESSION
			return open_cgi_session(req, conn, id);
		}
		case SVR_HTTP:
// HTTP SESSION
			return open_http_session(req, conn, id);
	}
	conn.result = DBGSE_WRONGREQ;
	return E_FAIL;
}


SessConnection::~SessConnection() {
	if (cgi_pid) {
		char buf[256];
		for (int cnt=0; receiver.recv(buf, sizeof(buf), 100) > 0 && cnt < 100; cnt++) {}
#ifdef WIN32
		HANDLE h = OpenProcess(PROCESS_TERMINATE, FALSE, cgi_pid);
		if (h) {
			TerminateProcess(h, 1);
			CloseHandle(h);
		}
#else
		kill(cgi_pid, SIGTERM);
#endif
	}
}

bool SessConnection::IsTerminated() {
	bool retval = IsCGI;
	if (cgi_pid && IsCGI) {
#ifdef WIN32
		HANDLE h = OpenProcess(PROCESS_TERMINATE, FALSE, cgi_pid);
		if (h) {
			retval = WaitForSingleObject(h, 1) == WAIT_OBJECT_0;
			if (retval) {
				DWORD uexitcode;
				GetExitCodeProcess(h, &uexitcode);
				exitcode=uexitcode;
			}
			CloseHandle(h);
		}
#else
		pid_t apid;
		int status;
		apid = waitpid(cgi_pid, &status, WNOHANG);
		retval = (apid < 0) || ((apid > 0) && WIFEXITED(status));
		if (retval) {
			exitcode = status;
		}
#endif
		if (retval) {
			cgi_pid = 0;
		}
	}
	return retval;
}
	
HRESULT SessConnection::assign_cgi(pid_t pid, pipe_t houtputpipe, pipe_t herrpipe) {
	cgi_pid = pid;
	if (pid == 0) {
		result = DBGSE_CGIHANDLERNOTFOUND;
    		return E_FAIL;
	}
	IsCGI = true;
	
	if (receiver.assignpipe(houtputpipe) != 0) {
		CLOSEPIPE(houtputpipe);
		CLOSEPIPE(herrpipe);				
		result = DBGSE_ASSIGNPIPE;
		return E_FAIL;
	}
	
	if (errreceiver.assignpipe(herrpipe) != 0) {
		CLOSEPIPE(herrpipe);
		receiver.close();
		result = DBGSE_ASSIGNPIPE;
		return E_FAIL;
	}
	return IsTerminated() ? E_FAIL : S_OK;
}
