/***************************************************************************
                          ListenerBase.cpp  -  description
                             -------------------
    begin                : Fri Mar 2 2001
    copyright            : (C) 2001 by Dmitri Dmitrienko
                         : (C) 2002, 2007 NuSphere Corp.
    www                  : http://dd.cron.ru
                         : http://www.nusphere.com/
    license              : This source file is subject to version 1.01 of 
                           the License,  that is bundled with this package 
                           in the file LICENSE, and is available at through 
                           the world-wide-web at http://dd.cron.ru/license
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#include "dbgbase/dbgbase_classes.h"

#include "DbgSess.h"

ListenerBase::~ListenerBase() {
	DBG_TRACE(("ListenerBase::~ListenerBase\n"));
	stop(true);
	th_list.clear();
}

void ListenerBase::stop(bool stopclients) {
	TH_LIST::iterator p;

	DBG_TRACE(("ListenerBase::stop\n"));
	stopJIT();
	if (stopclients) {
		DBG_TRACE(("stopping clients\n"));
		for (p = th_list.begin(); p!=th_list.end(); p++) {
			(*p).clientthread->stop(false);
		}
		for (p = th_list.begin(); p!=th_list.end(); p++) {
			(*p).clientthread->stop(true);
		}
		while (th_list.size() > 0) {
			p = th_list.begin();
			if ((*p).clientthread->isfinished()) {
				destroy_client((*p).clientthread);
				th_list.erase(p);
			}
		}
	}
}

// J I T   d e b u g g i n g   s u p p o r t


int ListenerBase::startJIT(cfgprm *cfg, void *extcfg) {
	DBG_TRACE(("ListenerBase::startJIT\n"));
	fcfgJIT = *cfg;
	
	if (fsocketJIT) {
		DBG_TRACE(("socket was not closed.\n"));
		DBG_TRACE(("dbh_close_socket(%d)\n", fsocketJIT));
		dbh_close_socket(fsocketJIT);
		fsocketJIT = 0;
	}
	fsocketJIT = dbh_create_listen_socket(fcfgJIT.address.c_str(), fcfgJIT.port);
	if (fsocketJIT <= 0) {
		ferrorJIT = oserrno(1); // get network error
		fstatusJIT = LSTNRS_initerror;
		switch (fsocketJIT) {
			case -1: 
				DBG_TRACE(("API socket()"));
				break;
			case -2:
				DBG_TRACE(("API bind()"));
				break;
			case -3:
				DBG_TRACE(("API listen()"));
				break;
			case -4:
				DBG_TRACE(("API gethostbyname()"));
				break;
			default:
				DBG_TRACE(("socket creation"));
		}
		DBG_TRACE((" failed (err=%d). ",ferrorJIT));
		fsocketJIT = 0;	
	} else {
		ferrorJIT = 0;
		fstatusJIT = LSTNRS_ok;
	}
	return fsocketJIT > 0;
}

void ListenerBase::stopJIT() {
	DBG_TRACE(("ListenerBase::stopJIT\n"));
	if (fsocketJIT > 0) {
		DBG_TRACE(("dbh_close_socket(%d)\n", fsocketJIT));
		dbh_close_socket(fsocketJIT);
		fsocketJIT = 0;
	}
	ferrorJIT = 0;
	fstatusJIT = LSTNRS_stopped;
}

int ListenerBase::restartJIT(cfgprm *newcfg, void *extcfg) {

	DBG_TRACE(("ListenerBase::restartJIT\n"));
	if (fsocketJIT <= 0 || (
		fcfgJIT.address.length() && 
		newcfg->address.length() &&	
		stricmp(fcfgJIT.address.c_str(), newcfg->address.c_str()) != 0)
		|| fcfgJIT.port!=newcfg->port) {
		stopJIT();
		return startJIT(newcfg, extcfg);
	} else {
		fcfgJIT = *newcfg;
		return fsocketJIT>0;
	}
}

HRESULT ListenerBase::peekJIT(int serversocket, int sesswait_timeout, int client_timeout, 
							  const char *SID, int SOF, 
							  ThreadBase *th, DbgSiteBase *site, SessConnection *conn) {
	int fclientsocketREQ;
	sockaddr client_addr;
	HRESULT ret_val = E_FAIL;
	dbg_packet pack;
	dbg_header_struct hdr;
	dbg_frame *frame;
	char *pSID = NULL;
	int sz, rv;
	dbg_sid_body *sid_body;
	bool eq, islocalth;
	thlistitem li;
	char *addrstr;

	if (serversocket == 0)
		serversocket = fsocketJIT;

	if (serversocket <= 0) 
		return E_FAIL;

	fclientsocketREQ = dbh_check_new_client_connect(serversocket, &client_addr, sesswait_timeout);
	
	if (fclientsocketREQ <= 0)
		return E_FAIL; // timeout

	DBG_TRACE(("peekJit(sock=%d, sesswt=%d, cliwt=%d, sid=%s, SOF=%d)\n", serversocket, sesswait_timeout, client_timeout, SON(SID), SOF));

	addrstr = inet_ntoa(((sockaddr_in *)&client_addr)->sin_addr);
	log(LSTNRL_msg, "%s client connected", addrstr);

	islocalth = false;
	if (!th) {
		th = create_client();
		islocalth = true;
	}

	li.clientthread = th;

	if (!th) {
		dbh_close_socket(fclientsocketREQ);
		return E_HANDLE;
	}


	dbg_packet_new(&pack);
	rv = dbh_wait_debug_event(&hdr, fclientsocketREQ, &pack, client_timeout);

	if (rv > 0 && hdr.cmd == DBGC_SID) {
		frame = dbg_packet_findfirstframe(&pack, FRAME_SID);
		if (frame) {
			sid_body = FRAME_DATA_PTR(dbg_sid_body, frame);
			dbg_packet_findstring(&pack, sid_body->isid, &pSID, &sz);
		}
	}

	eq = (!pSID || !SID || strcmp(SID, pSID) == 0);
	dbg_packet_free(&pack);

	ret_val = S_FALSE;
	if (eq) {
		if (!th->assign_connection(fclientsocketREQ, &client_addr, SOF, client_timeout, site, conn)) {
			DBG_TRACE(("can not establish a connection with %s", addrstr));
			ret_val = E_FAIL;
		} else {
			th_list.push_back(li);
			DBG_TRACE(("new client is added\n"));
			ret_val = S_OK;
		}
	}

	if (ret_val != S_OK) {
		if (islocalth) {
			th->close();
			delete th;
		}
		if (fclientsocketREQ > 0)
			dbh_close_socket(fclientsocketREQ);
	}

	return ret_val;
}
	
bool ListenerBase::cleanup() {
	TH_LIST::iterator p;

	for (p = th_list.begin(); p!=th_list.end(); p++) {
		if ((*p).clientthread->isfinished()) {
			destroy_client((*p).clientthread);
			th_list.erase(p);
			return true;
		}
	}
	return false;
}

void ListenerBase::logstatusJIT() {
	switch (fstatusJIT) {
		case LSTNRS_notinitialized:
			log(LSTNRL_warn,"JIT Listener is not initialized,yet.");
			break;
		case LSTNRS_initerror:
			logsyserror("JIT Listener is not started. %s", ferrorJIT);
			break;
		case LSTNRS_stopped:
			log(LSTNRL_warn,"JIT Listener has been stopped.");
			break;
		case LSTNRS_ok:
			log(LSTNRL_msg,"JIT Listener is working properly.");
			break;
		default:
			log(LSTNRL_warn,"unknown JIT status %d.", fstatusJIT);
	}	
}


// R E Q   d e b u g g i n g   s u p p o r t

HRESULT ListenerBase::startsession(SessRequest& sessionreq, DbgSiteBase *fsite, DBGS_RESULT *result) {
	ThreadBase *th;
	SessConnection *aconn = NULL;
	HRESULT hr = E_FAIL;
	ListenerStatus savstatus;
	unsigned short port;
	bool ok, useJITsocket;
	int fsocketREQ = 0;
	
	DBG_TRACE(("ListenerBase::startsession\n"));
	if (!result) return E_POINTER;

// 
	savstatus = fstatusJIT;

// create thread
	th = create_client();
	*result = DBGSE_FAILEDCREATETHREAD;
	ok = (th != NULL);

	port = sessionreq.port;
	if (port == 0) port = DEFAULT_REQ_PORT;
	
	useJITsocket = false;
// open socket
	while (ok) {
		useJITsocket = (fcfgJIT.port == port && fstatusJIT == LSTNRS_ok);
		if (useJITsocket) {
			DBG_TRACE(("reusing JIT socked for starting REQ session\n"));
			fsocketREQ = fsocketJIT;
			fsocketJIT = 0;
		} else {
			fsocketREQ = dbh_create_listen_socket(sessionreq.address.c_str(), port);
		}
		ok = (fsocketREQ > 0);
		if (ok) break;
		*result = DBGSE_FAILEDCREATESOCKET;
		switch (fsocketREQ) {
			case -2:
				*result = DBGSE_FAILEDBINDSOCKET;
				if (sessionreq.port == 0 && port < (DEFAULT_REQ_PORT + REQ_PORT_COUNT)) {
				    ok = true;
					DBG_TRACE(("selecting next port\n"));
				    port++;
					if (useJITsocket)
						fsocketJIT = fsocketREQ;
				}
				break;
			case -3:
				*result = DBGSE_FAILEDLISTENSOCKET;
				break;
		}
	}

	if (ok) {
		sessionreq.port = port;
		aconn = new SessConnection();
		*result = DBGSE_FAILED;
		ok = (aconn != NULL);
	}

// start CGI process/HTTP connection
	if (ok) {
		hr = open_dbg_session(sessionreq, *aconn);
		*result = aconn->result;
		ok = SUCCEEDED(hr);
	}

	if (ok) {
		do {
			int iterations = 1, wt = sessionreq.timeoutms_waitsess, i;

			if (aconn->IsCGI) {
				iterations = sessionreq.timeoutms_waitsess / 1000;
				if (iterations < 1) iterations = 1;
				wt = 1000;
			}
			DBG_TRACE(("Starting peekJIT\n"));
		
			for (i = 0; i < iterations; i++) {
				hr = peekJIT(fsocketREQ, wt, sessionreq.timeoutms_client, 
					aconn->sessid.c_str(), sessionreq.SOF_flags, th, fsite, aconn);
				DBG_TRACE(("peekJIT returned ->%d\n", hr));
				ok = (hr == S_OK);
				if (ok) break;
				
				if (aconn->IsTerminated()) {
					DBG_TRACE(("CGI child process is terminated, %d\n", aconn->exitcode));
					log(LSTNRL_error, "CGI child process is terminated, %d", aconn->exitcode);
					hr = E_FAIL;
					*result = DBGSE_TIMEOUTSTARTSESS;
					break;
				}
			}
			if (FAILED(hr)) {
				*result = DBGSE_TIMEOUTSTARTSESS;
				break;
			}
		} while (hr == S_FALSE); // Wrong SID received
	}

// close socket
	if (useJITsocket) {
		fsocketJIT = fsocketREQ;
		fsocketREQ = 0;
	}

	if (fsocketREQ > 0) {
		dbh_close_socket(fsocketREQ);
		fsocketREQ = 0;
	}
	
// finally ok
	if (ok) {
		*result = DBGSE_SUCCESSFULL;
		return S_OK;
	}
// finally fail
	if (th) {
		th->close();
		delete th;
	}

	if (fsite) 
		fsite->close();

	if (aconn) 
		delete aconn;

	return hr;
}

// M I S C   l o g g i n g   s e r v i c e 
void ListenerBase::logsyserror(const char* message, int err) {
	std::string buf, sbuf;

	if (!err) err = oserrno(0);
	if (!err) err = oserrno(1);
	if (!err) return;
	oserrnostr(buf, err);
	std_sprintf(sbuf, "OSERR (0x%X) '%s'", err, buf.c_str());
	log(LSTNRL_error, message, sbuf.c_str());
}

