/***************************************************************************
                          ParseEval.cpp  -  description
                             -------------------
    begin                : Fri Oct 19 2001
    copyright            : (C) 2001 by Dmitri Dmitrienko
                         : (C) 2002, 2007 NuSphere Corp.
    www                  : http://dd.cron.ru
                         : http://www.nusphere.com/
    license              : This source file is subject to version 1.01 of 
                           the License,  that is bundled with this package 
                           in the file LICENSE, and is available at through 
                           the world-wide-web at http://dd.cron.ru/license
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string>
#include <vector>
#include "dbgbase/dbgbase_globals.h"

char *valtypes[10] = {"undefined","long","double","string","array","object","bool","resource","reference","soft reference"};

inline bool ExtractSubStr(const std::string& str, unsigned int *startIdx, char chstart, char chend, std::string& rslt) {
	unsigned int idx=(*startIdx);
	if (idx >= (str.length() - 1) || str[idx] != chstart) return false;
	int i = ++idx;
	i = str.find(chend, i);
//	while (i < str.length() && str[i] != chend) {i++;}
	if (i==str.npos) return false;
	rslt.assign(str, idx, i-idx);
	//str.copy(, i-(*startIdx)-1, (*startIdx)+1);
	//rslt=str.substr((*startIdx)+1,i-(*startIdx)-1);
	*startIdx = i+1;
	return true;
}

inline bool ExtractQuotedSubStr(const std::string& str, unsigned int *startIdx, unsigned int slen, std::string *rslt) {	
	unsigned int idx=(*startIdx);
	if (idx+slen+1 >= str.length() || 
		str[idx] != '"' ||
		str[idx+slen+1] != '"') return false;
	unsigned int i=idx+1;
	rslt->assign(str, i, slen);
	(*startIdx)+=slen+2;
	return true;
}

inline bool ExtractInt(const std::string& str, unsigned int *startIdx, char chstart, char chend, long* rslt) {
	std::string subs;	
	char *eptr;
	if (!ExtractSubStr(str, startIdx, chstart, chend, subs)) return false;
	*rslt = strtol(subs.c_str(), &eptr, 10);
	return (eptr[0]=='\0');
}

inline void ConvertToPhpString(const std::string& s, std::string* result) {
	unsigned int i,ri,n,len;
	char ch;
	std::string sn;
	std::string rslt;
	
	len=s.length()*2 + 10;
	rslt.resize(len);
	rslt[0] = '"';
	ri=1;
	for (i=0; i < s.length(); i++) {
		ch=s[i];
		switch (ch) {
			case '"':
			case '\\':
			case '$': 
				rslt[ri] = '\\';
				rslt[ri+1] = ch;
				ri+=2;
				break;
			case '\r':
				rslt[ri]='\\';
				rslt[ri+1]='r';
				ri+=2;
				break;
			case '\n':
				rslt[ri]='\\';
				rslt[ri+1]='n';
				ri+=2;
				break;
			case 9:
				rslt[ri]='\\';
				rslt[ri+1]='t';
				ri+=2;
				break;
			default:
				if (ch>=0 && ch < 32) {
					rslt[ri]='\\';
					rslt[ri+1]='x';
					ri+=2;
					std_sprintf(sn, "%x", (int)ch);
					for (n=0; n < sn.length(); n++, ri++) {
						rslt[ri]=sn[n];
					}
				} else {
					rslt[ri]=ch;
					ri++;
				}
		}
		if (ri >= len-8) {
		  len+=128;
		  rslt.resize(len);
		}
	}
	rslt[ri] = '"';
	rslt.resize(ri+1);
	result->assign(rslt);
}




bool ParseEvalFull(const std::string &name, EVALLIST* list, VARLIST* var_list, const std::string& str, unsigned int *startIdx, bool MakePhpStr);

inline bool ParseEvalNULL(const std::string &name, EVALLIST* list, VARLIST* var_list, const std::string& str, unsigned int *startIdx) {
	unsigned int idx=*startIdx;
	if (idx >= str.length() || str[idx] != ';') return false;
	(*startIdx)++;
	list->Add(var_list, name, "NULL");
	return true;
}

inline bool ParseEvalInt(const std::string &name, EVALLIST* list, VARLIST* var_list, const std::string& str, unsigned int *startIdx) {
	std::string subs;
	if (!ExtractSubStr(str,startIdx,':',';',subs)) return false;
	list->Add(var_list, name, subs, ptLong);
	return true;
}

inline bool ParseEvalDouble(const std::string &name, EVALLIST* list, VARLIST* var_list, const std::string& str, unsigned int *startIdx) {
	std::string subs;
	if (!ExtractSubStr(str,startIdx,':',';',subs)) return false;
	list->Add(var_list, name, subs, ptDouble);
	return true;
}

inline bool ParseEvalString(const std::string &name, EVALLIST* list, VARLIST* var_list, const std::string& str, unsigned int *startIdx, bool MakePhpStr) {
	long slen;
	if (!ExtractInt(str, startIdx, ':', ':', &slen)) return false;
	if (*startIdx >= str.length()-1 || str[*startIdx] != '"') return false;
	(*startIdx)++;
	std::string subs = str.substr(*startIdx,slen);
	(*startIdx)+=slen;
	if ((*startIdx) >= str.length()-1 || str[*startIdx]!='"' || str[(*startIdx)+1]!=';' || (int)subs.length()!=slen) return false;  
	(*startIdx)+=2;
	if (MakePhpStr) {
		ConvertToPhpString(subs, &subs);
	}
	list->Add(var_list, name, subs, ptString);
	return true;
}

inline bool ParseEvalArray(const std::string &name, EVALLIST* list, VARLIST* var_list, const std::string& str, unsigned int *startIdx, const std::string& classname, php_type atype) {
	long arritems;
	
	std::string itemname;

	if (!ExtractInt(str, startIdx, ':', ':', &arritems)) return false;
	if (*startIdx >= str.length()-1 || str[*startIdx]!='{') return false;
	(*startIdx)++;

	eval_item *item = list->Add(var_list, name, classname, atype);
	if (arritems > 0) {
		item->subitems = new EVALLIST();
	} else if (str[*startIdx]!='}') return false;

	while (*startIdx < str.length() && str[*startIdx]!='}') {
		EVALLIST tmplst;
		// name
		if (!ParseEvalFull("", &tmplst, NULL, str, startIdx, false) || tmplst.size()!=1) return false;
		// value
		if (!ParseEvalFull(tmplst[0]->value, item->subitems, var_list, str, startIdx, true)) return false;
	}
	(*startIdx)++;
	return true;
}

inline bool ParseEvalObject(const std::string &name, EVALLIST* list, VARLIST* var_list, const std::string& str, unsigned int *startIdx) {
	long slen;
	std::string classname;

	if (!ExtractInt(str, startIdx, ':', ':', &slen)) return false;
	if (*startIdx >= str.length()-1) return false;
	if (!ExtractQuotedSubStr(str, startIdx, slen, &classname)) return false;
	if ((int)classname.length()!=slen) return false;
	return ParseEvalArray(name, list, var_list, str, startIdx, classname, ptObject);
}


inline bool ParseEvalBool(const std::string &name, EVALLIST* list, VARLIST* var_list, const std::string& str, unsigned int *startIdx) {
	long v;
	if (!ExtractInt(str, startIdx, ':', ';', &v)) return false;
	list->Add(var_list, name, (v==0)?("FALSE"):("TRUE"), ptBool);
	return true;
}


inline bool ParseEvalResource(const std::string &name, EVALLIST* list, VARLIST* var_list, const std::string& str, unsigned int *startIdx) {
	long v, slen;
	std::string restype, val;

	if (!ExtractInt(str, startIdx, ':', ':', &slen)) return false;
	if (!ExtractQuotedSubStr(str, startIdx, slen, &restype)) return false;
	if (!ExtractInt(str, startIdx, ':', ';', &v)) return false;
	std_sprintf(val, "%ld", v);
	list->Add(var_list, name, val, ptResource, restype);
	return true;
}

inline bool ParseEvalRef(const std::string &name, EVALLIST* list, VARLIST* var_list, const std::string& str, unsigned int *startIdx, bool isSoftRef) {
	long v;

	if (!ExtractInt(str, startIdx, ':', ';', &v)) return false;

	peval_item item=list->Add(NULL, name, "", (isSoftRef)? (ptSoftRef) : (ptRef));
	v--; // ref ID is 1-based, EvalList is 0-based
	item->var_index = v;
	if (!var_list || v < 0 || v >= (int)var_list->size()) {
		item->ref = item; // self-resolving
		return true;
	}	
	item->ref = (*var_list)[v];
	return true;
}


bool ParseEvalFull(const std::string &name, EVALLIST* list, VARLIST* var_list, const std::string& str, unsigned int *startIdx, bool MakePhpStr) {
	bool ret_val = false;

	if (*startIdx >= str.length()) return false;
	char ch = str[*startIdx];
	(*startIdx)++;
	switch (ch) {
		case 'N': 
			ret_val = ParseEvalNULL(name, list, var_list, str, startIdx);
			break;
		case 'i': 
			ret_val = ParseEvalInt(name, list, var_list, str, startIdx);
			break;
		case 'd': 
			ret_val = ParseEvalDouble(name, list, var_list, str, startIdx);
			break;
		case 's': 
			ret_val = ParseEvalString(name, list, var_list, str, startIdx, MakePhpStr);
			break;
		case 'a': 
			ret_val = ParseEvalArray(name, list, var_list, str, startIdx, "", ptArray);
			break;
		case 'O': 
			ret_val = ParseEvalObject(name, list, var_list, str, startIdx);
			break;
		case 'b': 
			ret_val = ParseEvalBool(name, list, var_list, str, startIdx);
			break;
		case 'z': 
			ret_val = ParseEvalResource(name, list, var_list, str, startIdx);
			break;
		case 'R': 
			ret_val = ParseEvalRef(name, list, var_list, str, startIdx, false);
			break;
		case 'r': 
			ret_val = ParseEvalRef(name, list, var_list, str, startIdx, true);
			break;
	}
	if (!ret_val) { // try to recover
		unsigned int i=*startIdx;
		while (i<str.length() && str[i]!='{' && str[i]!=';') i++;
		if (i<str.length() && str[i] == '{') {
			unsigned int cnt=1;
			i++;
			while (i<str.length() && cnt!=0) {
				if (str[i] == '{')
					cnt++;
				else if (str[i] == '}') 
					cnt--;
				i++;
			}
		}
		*startIdx = i;
	}

	return 	ret_val;
}

EVALLIST::EVALLIST(const std::string& serial_str, const std::string& name) {
	VARLIST var_list;
/*	__int64 T1,T2, T, C; 
	double v;
	
	QueryPerformanceCounter((_LARGE_INTEGER *)&T1);*/
	unsigned int startIdx = 0;
	ParseEvalFull(name, this, &var_list, serial_str, &startIdx, true);
/*	QueryPerformanceCounter((_LARGE_INTEGER *)&T2);
	T=T2-T1;
	QueryPerformanceFrequency((_LARGE_INTEGER *)&C);
	v=T;
	v=v/C;
	printf("\nT=%g\n",v*1000.0);*/
};
