/***************************************************************************
                          PipeReader.cpp  -  description
                             -------------------
    begin                : Fri Mar 2 2001
    copyright            : (C) 2001 by Dmitri Dmitrienko
                         : (C) 2002, 2007 NuSphere Corp.
    www                  : http://dd.cron.ru
                         : http://www.nusphere.com/
    license              : This source file is subject to version 1.01 of 
                           the License,  that is bundled with this package 
                           in the file LICENSE, and is available at through 
                           the world-wide-web at http://dd.cron.ru/license
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#if HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_SYS_TYPES
#include <sys/types.h>
#endif

#include <assert.h>
#include "dbgbase/dbgbase_classes.h"

PipeReader::PipeReader() :
	fpipe(INVALID_PIPE), fsocket(0) {
}


int PipeReader::assignpipe(pipe_t apipe) {
	if (fpipe != INVALID_PIPE || fsocket != 0) return (-1);
	fpipe = apipe;
	return 0;
}

int PipeReader::assignsocket(int asocket) {
	if (fpipe != INVALID_PIPE || fsocket != 0) return (-1);
	fsocket = asocket;
	return 0;
}

void PipeReader::close() {
	CLOSEPIPE(fpipe);
	if (fsocket) {
		::SCLOSE(fsocket);
	}
}

int sockrecv(int socket, int is_socket, char *abuf, int len, int timeoutms) {
		fd_set rset;
		struct timeval timeout;
		int ret_val;

		FD_ZERO(&rset);
		FD_SET((unsigned int) socket, &rset);
		if (timeoutms >= 0) {
			timeout.tv_sec = timeoutms / 1000;
			timeout.tv_usec = (timeoutms>1000) ? (0) : (timeoutms * 1000);
		} else {
			timeout.tv_sec = 0;
			timeout.tv_usec = 0;
		}
		ret_val = select(socket + 1, &rset, NULL, NULL, &timeout);
		
		if (ret_val != 1) {	
			return ((ret_val < 0) ? -1:0);
		}

		if (is_socket) {
			ret_val = ::recv(socket, abuf, len, 0);
		} else {
			ret_val = ::read(socket, abuf, len);
		}
		return (ret_val <= 0) ? (-1) : (ret_val);
}

int PipeReader::recv(char *abuf, int len, int timeoutms) {
	DBG_TRACE(("PipeReader::recv(buf=%0.8x, len=%d)\n", (int)abuf, len));
	if (fsocket != 0) {
		return sockrecv(fsocket, 1, abuf, len, timeoutms);
	}
	if (fpipe != INVALID_PIPE) {
		return sockrecv(fpipe, 0, abuf, len, timeoutms);
	}
	return -2; // uninitialized PipeReader
}

