/***************************************************************************
                          StrTable.cpp  -  description
                             -------------------
    begin                : Fri Mar 2 2001
    copyright            : (C) 2001 by Dmitri Dmitrienko
                         : (C) 2002, 2007 NuSphere Corp.
    www                  : http://dd.cron.ru
                         : http://www.nusphere.com/
    license              : This source file is subject to version 1.01 of 
                           the License,  that is bundled with this package 
                           in the file LICENSE, and is available at through 
                           the world-wide-web at http://dd.cron.ru/license
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#include "StrTable.h"

const char *MERGE_APPEND_CHAR[] = {"","",",",";",":"};

std::string& trimleft(std::string& str) {
	unsigned int pos;
	for (pos=0; pos < str.length() && (str[pos]==' ' || str[pos]=='\t'); pos++);
	if (pos > 0) str.erase(0, pos);
	return str;
}

std::string& trimright(std::string& str) {
	unsigned int pos;
	for (pos=str.length() - 1; pos >= 0 && (str[pos]==' ' || str[pos]=='\t'); pos--);
	if (pos < str.length()) str.erase(pos+1, str.length());
	return str;
}

std::string& trimstr(std::string& str) {
	return trimright(trimleft(str));
}


// StringTable

StringTable::StringTable(const std::string& str, char equ_delimiter, const char *line_delimiter) {
	assign(str, equ_delimiter, line_delimiter);
}

StringTable::StringTable(char** strarray, char equ_delimiter) {
	assign(strarray, equ_delimiter);
}

void StringTable::assign(char** strarray, char equ_delimiter) {
	char const* pline;
	int equ_pos;
	std::string line, key, val;
	for (; (pline = *strarray); strarray++) {
		line = pline;
		if (line.length() > 0) {
			equ_pos = line.find(equ_delimiter);
			key = line.substr(0, equ_pos);
			trimright(key);
			val = line.substr(equ_pos + 1, line.length());
			trimleft(val);
			add(key, val);
		}
	}
}

void StringTable::assign(const std::string& str, char equ_delimiter, const char *line_delimiter) {
	int pos, equ_pos, delim_len = strlen(line_delimiter);
	unsigned int hd;
	std::string line, key, val;

	clear();
	hd = 0;
	while (hd <= str.length()) {
		pos = str.find(line_delimiter, hd);
		if (pos == -1) pos = str.length();
		line = str.substr(hd, pos - hd);
		if (line.length() > 0) {
			equ_pos = line.find(equ_delimiter);
			key = line.substr(0, equ_pos);
			trimright(key);
			val = line.substr(equ_pos + 1, line.length());
			trimleft(val);
			add(key, val);
		}
		hd = pos + delim_len;
	};
}


void StringTable::merge(const std::string& key, const std::string& val, MERGETYPE flags) {	
	iterator it;
	const char *ap;
	for (it = begin(); it!=end(); it++) {
		if (it->key == key) {
// merging rules
			switch (flags) {
				case MERGE_OVERWRITE:
					it->val = val;
					break;
				case MERGE_APPEND_W_COMMA:
				case MERGE_APPEND_W_SEMICOLON:
				case MERGE_APPEND_W_COLON:
					ap = MERGE_APPEND_CHAR[flags & MERGE_APPEND_MASK];
					it->val.append(ap).append(val);
			}
			return;
		}
	}
	add(key, val);
}

void StringTable::merge(const StringTable& atable, MERGETYPE flags) {
	const_iterator it;
	for (it = atable.begin(); it!=atable.end(); it++) {
		merge(it->key, it->val, flags);
	}
}

void StringTable::getline(int index, std::string& str, const char *delimiter, const char *line_delimiter) const {
	const_iterator it;
	if (index < 0 || index >= (int)size()) {
		str = "";
		return;
	}
	it = begin() + index;
	if (it->key.length() == 0 && it->val.length() == 0) {

		str = line_delimiter;
	} else if (it->key.length() == 0 || it->val.length() == 0) {
		str = it->key + it->val + line_delimiter;
	} else {
		str = it->key + delimiter + it->val + line_delimiter;
	}
}


void StringTable::getlines(std::string& str, const char *delimiter, const char *line_delimiter) const {
	int i;
	std::string s;
	str = "";
	for (i = 0; i < (int)size(); i++) {
		getline(i, s, delimiter, line_delimiter);
		str.append(s);
	}
}

// SimpleStringArray

SimpleStringArray::SimpleStringArray(const StringTable& table, char *equ_delimiter) {
//	StringTableBase::const_iterator it;
	std::string str;
	char **pp;

	size = table.size()+1;
	farray = (char**)calloc(size * sizeof(char*), 1);
	
	pp = farray;
	for (int i = 0; i < (int)table.size(); i++) {
		table.getline(i, str, equ_delimiter, "");
		*pp = strdup(str.c_str());
		pp++;
	}
/*	for (it = table.begin(); it!=table.end(); it++) {
		str = it->key;
		str.append(equ_delimiter).append(it->val);
		*pp = strdup(str.c_str());
		pp++;
	}*/
}

SimpleStringArray::SimpleStringArray(const SimpleStringArray& src) : farray(NULL) {
	char **pp;
	char **ppsrc;

	size = src.count();
	farray = (char**)calloc((size + 1) * sizeof(char*), 1);

	pp = farray;
	
	for (ppsrc = src.farray; ppsrc && *ppsrc; ppsrc++, pp++) {
		*pp = strdup(*ppsrc);
	}
}

void SimpleStringArray::add(const char *str) {
	size++;
	farray = (char**)realloc(farray, (size + 1) * sizeof(char*));
	farray[size-1] = strdup(str);
	farray[size] = NULL;
}

void SimpleStringArray::clear() {
	char **pp;
	if (!farray) return;
	pp = farray;
	for (; *pp; pp++) {
		free(*pp);
	}
	free(farray);
	farray = NULL;
	size = 0;
}

SimpleStringArray& SimpleStringArray::operator = (const SimpleStringArray& src) {
	char **ppsrc, **pp;
	clear();
	size = src.count();
	if (size) {
		farray = (char**)calloc((size + 1) * sizeof(char*), 1);
		ppsrc = src.farray;
		pp = farray;
		if (ppsrc) {
			for (; *ppsrc; ppsrc++, pp++) {
				*pp = strdup(*ppsrc);
			}
		}
	}
	return (*this);
}

