/***************************************************************************
                          ThreadBase.cpp  -  description
                             -------------------
    begin                : Fri Mar 2 2001
    copyright            : (C) 2001 by Dmitri Dmitrienko
                         : (C) 2002, 2007 NuSphere Corp.
    www                  : http://dd.cron.ru
                         : http://www.nusphere.com/
    license              : This source file is subject to version 1.01 of 
                           the License,  that is bundled with this package 
                           in the file LICENSE, and is available at through 
                           the world-wide-web at http://dd.cron.ru/license
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#if HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include "dbgbase/dbgbase_classes.h"

#define SERVER_INWAIT (server_flags & DBGF_WAITACK)
#define WAIT_GRANUL 300

bool ThreadBase::assign_connection(
	int sock, 
	sockaddr *client_addr, 
	dbgint aSOF_flags, 
	int timeoutms_client, 
	DbgSiteBase *asite, 
	SessConnection *aconn) 
{
	DBG_TRACE(("ThreadBase::assign_connection(%d, %d, %p)\n", sock, aSOF_flags, asite));
	if (fsock > 0 || fsite != NULL || fconn != NULL) return false; // smth wrong.
	fsite = asite;
	fconn = aconn;
	if (fconn) {
		fconnbuf = new char[CONNBUFSIZE];
		fconnerrbuf = new char[CONNERRBUFSIZE];
	}
	fsock = sock;
	fclient_addr = *client_addr;
	SOF_flags = aSOF_flags;
	ftimeoutms_client = timeoutms_client;
	return true;
}
	
void ThreadBase::close() {
	DBG_TRACE(("ThreadBase::close()\n"));
	if (fsock) {
		DBG_TRACE(("closing socket %d\n",fsock));
		dbh_close_socket(fsock);
		fsock = 0;
	}
	scopelist.clear();
	breakpointlist.clear();
	if (fsite) {
		if (siteinitialized) {
			siteinitialized = 0;
			fsite->close();
		}
		if (fsite) {
			delete fsite;
			fsite = NULL;
		}
	}
	if (fconn) {
		delete fconn;
		fconn = NULL;
	}
	if (fconnbuf) {
		delete fconnbuf;
		fconnbuf = NULL;
	}
	if (fconnerrbuf) {
		delete fconnerrbuf;
		fconnerrbuf = NULL;
	}

	flags &= ~DBGF_WAITACK;
	server_flags = 0;
}

//////////////// PROCESS() PART ////////////////////

void ThreadBase::process_conn() {
	DBG_TRACE(("ThreadBase::process_conn>\n"));
	if (fconn) {
		while (fconn->receiver.recv(fconnbuf, CONNBUFSIZE, 0) > 0) {};
		while (fconn->errreceiver.recv(fconnerrbuf, CONNERRBUFSIZE, 0) > 0) {};
	}
	DBG_TRACE(("ThreadBase::process_conn<\n"));
}

int ThreadBase::check_site_alive() {
	HRESULT hr;
	int ret_val = 0;

	if (fsite) {
		hr = fsite->isalive();
		if (hr==S_FALSE) {
			DBG_TRACE(("PAUSE REQ.\n"));
			ret_val = dbg_packet_send(DBGC_PAUSE, NULL, fsock, 0);
			if (ret_val > 0) ret_val = 0;
		} else {
			if (hr!=S_OK) {
				DBG_TRACE(("TERM REQ.\n"));
				ret_val = -1;
			}
		}
	}
	return ret_val;
}

void ThreadBase::process() {
	dbg_packet pack;
	int ret_val = 0;
	dbg_header_struct hdr;	
	
	DBG_TRACE(("ThreadBase::process>\n"));
	if (SERVER_INWAIT) {
		DBG_TRACE(("Unexpected INWAIT\n"));
		ret_val = ack(DBGA_CONTINUE, NULL);
		if (ret_val < 0) return;
	}
	
	process_conn();
	
	if (fsock == 0) {
	    DBG_TRACE(("Unexpected fsock==0\n"));
	    return;
	}

	dbg_packet_new(&pack);
	int time_waited = 0;
	do {
		ret_val = dbh_wait_debug_event(&hdr, fsock, &pack, WAIT_GRANUL);
		time_waited+=WAIT_GRANUL;
		if (ret_val == 0) { // check whether the client wants to stop
			process_conn();
			ret_val = check_site_alive();
		} else if (ret_val < 0) {
			DBG_TRACE(("failed recv returned %d, errno=%d.\n",ret_val,errno));
		}
	} while (ret_val == 0 && time_waited < ftimeoutms_client);

	if (ret_val > 0) {
		server_flags = hdr.flags;
		ret_val = process_cmd(hdr.cmd, &pack);
	}
	if (ret_val < 0 || SERVER_INWAIT) {
		if (SERVER_INWAIT) {
			DBG_TRACE(("STOP.\n"));
			ack(DBGA_STOP, NULL);
		}
		close();
	}
	DBG_TRACE(("freing pack\n"));
	dbg_packet_free(&pack);
	DBG_TRACE(("ThreadBase::process<\n"));
}

int ThreadBase::process_cmd(int cmd, dbg_packet *pack) {
	BR_REASON reason = (BR_REASON)-1;
	int ret_val;

	DBG_TRACE(("ThreadBase::process_cmd(%d)>\n", cmd));

	scopelist.assign(pack);

	if (cmd == DBGC_STARTUP) {
		if (flags & DBGF_STARTED) return -1; /* already started */
		setsessionoptions(SOF_flags);
		ret_val = onstartsession(pack);
		DBG_TRACE(("ThreadBase::process_cmd(ret=%d)<\n", ret_val));
		return ret_val;
	}

/* all other commands can be processed while session is started */

	if (!(flags & DBGF_STARTED) || fsite == NULL) {
		DBG_TRACE(("site is not initialized yet\n"));
		return 0;
	}

	switch (cmd) {
		case DBGC_END:
			return onendsession(pack);
		case DBGC_ERROR:
			return onerror(pack);
		case DBGC_LOG:
			return onlog(pack);
		/* breakpoints */
		case DBGC_BREAKPOINT:
			reason = BR_BREAKPOINT;
			break;
		case DBGC_EMBEDDED_BREAK:
			reason = BR_EMBEDDED;
			break;
		case DBGC_STEPINTO_DONE:
			reason = BR_STEPINTO;
			break;
		case DBGC_STEPOVER_DONE:
			reason = BR_STEPOVER;
			break;
		case DBGC_STEPOUT_DONE:
			reason = BR_STEPOUT;
			break;
	}
	if (reason == (BR_REASON)-1) {
		DBG_TRACE(("unknown reason\n"));
		return -1;
	}

	return onbreakpoint(pack, reason);
}

char ThreadBase::isfinished() {
	return ((flags & DBGF_FINISHED)!=0);
}


/////////////////////// ACK /////////////////////////////

int ThreadBase::ack(dbgint cmd, dbg_packet *pack) {
	int ret_val;

	DBG_TRACE(("ThreadBase::ack(%d)\n", (int)cmd));
	process_conn();

	if (!SERVER_INWAIT) return 0;

	DBG_TRACE(("fdoclist.clear()\n"));
	fdoclist.clear();

	flags &= ~DBGF_WAITACK;
	server_flags &= ~DBGF_WAITACK;

	DBG_TRACE(("calling dbh_ack>\n"));
	ret_val = dbh_ack(cmd, pack, fsock, flags);
	DBG_TRACE(("calling dbh_ack<\n"));
	return ret_val;
}

int ThreadBase::ack_action(BP_ACTION action) {
	int ret_cmd = DBGA_STOP;
	int ret_val;

	DBG_TRACE(("ThreadBase::ack_action(%d)\n", (int)action));
	switch (action) {
		case BP_ABORT:
			ret_cmd = DBGA_STOP;
			break;
		case BP_STEP_INTO:
			ret_cmd = DBGA_STEPINTO;
			break;
		case BP_STEP_OVER:
			ret_cmd = DBGA_STEPOVER;
			break;
		case BP_STEP_OUT:
			ret_cmd = DBGA_STEPOUT;
			break;
		case BP_DELAY_HANDLING:
			ret_cmd = 0;
			break;
		default:
			ret_cmd = DBGA_CONTINUE;
	}
	ret_val = 0;
	if (ret_cmd) {
		ret_val = ack(ret_cmd, NULL);
		ret_val = (ret_val) ? 1:0;
	}
	return ret_val;
}

int ThreadBase::request(dbg_header_struct *hdr, dbg_packet *pack) {
	int ret_val;

	process_conn();

	if (!SERVER_INWAIT) {
		memset(hdr, 0, sizeof(dbg_header_struct));
		dbg_packet_clear(pack);
		return 0;
	}

	flags |= DBGF_WAITACK;
	server_flags &= ~DBGF_WAITACK;

	ret_val = dbh_ack(DBGA_REQUEST, pack, fsock, flags);
	if (ret_val > 0) {
		ret_val = dbg_packet_recv(hdr, pack, fsock, ftimeoutms_client);
		if (ret_val > 0) server_flags = hdr->flags;
	} else {
		dbg_packet_clear(pack);
	}
	if (ret_val < 0 || fsock <= 0) server_flags = DBGF_FINISHED;
	flags &= ~DBGF_WAITACK;
	return ret_val;
}

////////////// ON_XXXXXX METHODS ////////////////////

int ThreadBase::onstartsession(dbg_packet *pack) {
	dbg_version_body *ver_body;
	dbg_frame *frame;
	dbgint dbg_ver = 0x0;
	char *dbg_descr = NULL;
	int sz;
	int ret_val;
	std::string sdbg_descr;
	HRESULT hr;
	
	DBG_TRACE(("ThreadBase::onstartsession>\n"));	
	flags |= DBGF_STARTED;

	frame = dbg_packet_findfirstframe(pack,FRAME_VER);
	if (frame!=NULL) {
		ver_body = FRAME_DATA_PTR(dbg_version_body,frame);
		dbg_ver = ((ver_body->major_version & 0xFF) << 8) | (ver_body->minor_version & 0xFF);
		dbg_packet_findstring(pack, ver_body->idescription, &dbg_descr, &sz);
	} else {
		if (SUCCEEDED(getdbgversion(&dbg_ver, &sdbg_descr))) {
			dbg_descr = &sdbg_descr[0];
		}
	}

	if (fsite == NULL) {
		DBG_TRACE(("creating site\n"));	
		fsite = create_site();
	}
	if (!siteinitialized) {
		hr = fsite ? (fsite->initialize(this, dbg_ver, dbg_descr)) : E_FAIL;
		if (FAILED(hr)) {
			flags &= ~DBGF_STARTED;
			ack(DBGA_STOP, NULL);
//			close(); // close later in the loop
			return 0;
		}
		if (hr != S_OK) {
			flags &= ~DBGF_STARTED;
			ack(DBGA_CONTINUE, NULL);
//			close(); // close later in the loop
			return 0;
		}
		siteinitialized = 1;
	}

	if (SOF_flags & SOF_BREAKONLOAD) {
		ret_val = onbreakpoint(pack, BR_START_SESSION);
	} else {
		ret_val = ack(DBGA_CONTINUE, NULL);
	}

	DBG_TRACE(("ThreadBase::onstartsession<\n"));	
	return ret_val;
}

int ThreadBase::onendsession(dbg_packet *pack) {
	int ret_val;

	if (SOF_flags & SOF_BREAKONFINISH) {
		ret_val = onbreakpoint(pack, BR_END_SESSION);
	} else {
		ret_val = ack(DBGA_CONTINUE, NULL);
	}

	flags &= ~DBGF_STARTED;
	close();
	return ret_val;
}

int ThreadBase::onerror(dbg_packet *pack) {
	dbg_error_body *err;
	char *errmsg, *str;
	int siz, ret_val;
	BP_ACTION action = BP_CONTINUE;
	HRESULT hr;
	dbg_frame *frame;
	dbg_bpl_body *body;

	frame = dbg_packet_findfirstframe(pack, FRAME_BPL);
	if (frame!=NULL) {
		body = FRAME_DATA_PTR(dbg_bpl_body, frame);
		breakpointlist.assign(pack);
		is_valid_bplist = 1;
	}

	frame = dbg_packet_findfirstframe(pack, FRAME_ERROR);
	if (!frame) return 0;
	err = FRAME_DATA_PTR(dbg_error_body, frame);
	errmsg = NULL;
	if (dbg_packet_findstring(pack, err->imessage, &str, &siz)) {
		errmsg = str;
	}

	while (1) {
		hr = fsite->handleerror(err->type, errmsg, &action);
		if (action!=BP_DELAY_HANDLING || !SUCCEEDED(hr)) break;
		hr = fsite->delay_handling();
		if (!SUCCEEDED(hr)) break;
	}

	if (!SUCCEEDED(hr)) 
		action = BP_ABORT;
	ret_val = ack_action(action);
	return ret_val;
}

int ThreadBase::onbreakpoint(dbg_packet *pack, BR_REASON reason) {
	int ret_val;
	BP_ACTION action = BP_CONTINUE;
	HRESULT hr;
	dbg_frame *frame;
	dbg_bpl_body *body;

	DBG_TRACE(("ThreadBase::onbreakpoint(%d)\n", (int)reason));
	
	frame = dbg_packet_findfirstframe(pack, FRAME_BPL);
	if (frame!=NULL) {
		body = FRAME_DATA_PTR(dbg_bpl_body, frame);
		breakpointlist.assign(pack);
		is_valid_bplist = 1;
	}

	DBG_TRACE(("defer to site, %d\n", (int)fsite));

	while (1) {
		hr = fsite->handlebreakpoint(reason, &action);
		DBG_TRACE(("returned from site\n"));
		if (action!=BP_DELAY_HANDLING || !SUCCEEDED(hr)) break;
		hr = fsite->delay_handling();
		if (!SUCCEEDED(hr)) break;
	}

	if (!SUCCEEDED(hr)) 
		action = BP_ABORT; 
	else if (reason==BR_END_SESSION) 
		action = BP_CONTINUE;
	ret_val = ack_action(action);
	DBG_TRACE(("ThreadBase::onbreakpoint ret=%d<\n", ret_val));
	return ret_val;
}

int ThreadBase::onlog(dbg_packet *pack) {
	dbg_log_body *body;
	dbg_frame *frame;
	char *logmsg, *mod_name;
	int sz, szlog;

	if (!fsite) return 0;

	frame = dbg_packet_findfirstframe(pack,FRAME_LOG);
	while (frame) {
		body = FRAME_DATA_PTR(dbg_log_body, frame);
		dbg_packet_findstring(pack, body->ilog, &logmsg, &szlog);
		dbg_packet_findstring(pack, body->imod_name, &mod_name, &sz);
		fsite->log(body->mod_no, body->line_no, mod_name, body->type | (body->ext_info << 16), logmsg, szlog - 1);
		frame = dbg_packet_findnextframe(pack, FRAME_LOG, frame);
		process_conn();
	}
	check_site_alive();
	return 1;
}

//////////////// RequestorBase PART ////////////////////


HRESULT ThreadBase::isalive() {
	fd_set rset;
	timeval timeout;
	int ret_val;

	if (fsock) {
		FD_ZERO(&rset);
		FD_SET(fsock, &rset);
		timeout.tv_sec = 0;
		timeout.tv_usec = 0;
		ret_val = select(fsock + 1, &rset, NULL, NULL, &timeout);
		if (ret_val<0) return S_FALSE;
	}
	return (fsock<=0) ? S_FALSE : S_OK;
}

HRESULT ThreadBase::getdbgversion(
	/* [out] */ dbgint *dbg_ver,
	/* [out] */ std::string *sdbg_descr) {
	dbg_packet pack;
	dbg_header_struct hdr;
	int ret_val, sz;
	dbg_version_body lstnr_ver, *body;
	dbg_frame *frame;
	char *dbg_descr;
	HRESULT hr;

	DBG_TRACE(("ThreadBase::getdbgversion\n"));
	if (!sdbg_descr && !dbg_ver) return E_POINTER;

	if (!SERVER_INWAIT) return E_FAIL;

	*sdbg_descr = "";
	*dbg_ver = 0x0;

	dbg_packet_new(&pack);

	lstnr_ver.major_version = DBG_API_MAJOR_VERSION;
	lstnr_ver.minor_version = DBG_API_MINOR_VERSION;
	lstnr_ver.idescription = dbg_packet_add_string(&pack, DBG_API_DESCRIPTION);
	dbg_packet_add_frame(&pack, FRAME_VER, &lstnr_ver, sizeof(lstnr_ver));
	
	hr = E_FAIL;
	ret_val = request(&hdr, &pack);
	if (ret_val > 0) {
		frame = dbg_packet_findfirstframe(&pack,FRAME_VER);
		if (frame!=NULL) {
			body = FRAME_DATA_PTR(dbg_version_body,frame);
			*dbg_ver = ((body->major_version & 0xFF) << 8) | (body->minor_version & 0xFF);
			if (dbg_packet_findstring(&pack, body->idescription, &dbg_descr, &sz)) {
				*sdbg_descr = dbg_descr;
			}
			hr = S_OK;
		}
	}

	dbg_packet_free(&pack);
	return hr;
}

HRESULT ThreadBase::getserveraddr(
		/* [out] */ sockaddr_in *addr) {
	if (!addr) return E_POINTER;
	*addr = *(sockaddr_in *) &fclient_addr;
	return S_OK;
}

HRESULT ThreadBase::getdoclist(
		/* [in] */ DOCLIST *doclist) {
	dbg_packet pack;
	dbg_header_struct hdr;
	int ret_val = 0;

	if (!doclist) return E_POINTER;

	DBG_TRACE(("ThreadBase::getdoclist\n"));
	if (fdoclist.size()==0) {
		if (!SERVER_INWAIT) return E_FAIL;
		
		fdoclist.clear();
		dbg_packet_new(&pack);
		dbg_packet_add_frame(&pack, FRAME_SRC_TREE, NULL, 0);
		ret_val = request(&hdr, &pack);
		if (ret_val>0) {
			fdoclist.assign(&pack);
		}
		dbg_packet_free(&pack);
	} else {
		ret_val = 1;
	}
	if (fdoclist.size()==0) return E_FAIL;
	if (doclist!=&fdoclist) doclist->assign(fdoclist);
	return (ret_val > 0) ? S_OK : E_FAIL;
}

HRESULT ThreadBase::getlinesinfolist(
        /* [in] */ dbgint mod_no,
		/* [in] */ LINESINFOLIST *linesinfolist) {
	dbg_srclinesinfo_request req;	
	dbg_packet pack;
	dbg_header_struct hdr;
	int ret_val;

	DBG_TRACE(("ThreadBase::getlinesdata\n"));
	if (!linesinfolist) return E_POINTER;
	
	if (!SERVER_INWAIT) return E_FAIL;
	
	linesinfolist->clear();
	dbg_packet_new(&pack);
	req.mod_no = mod_no;
	dbg_packet_add_frame(&pack, FRAME_SRCLINESINFO, &req, sizeof(req));
	ret_val = request(&hdr, &pack);
	if (ret_val>0) {
		linesinfolist->assign(&pack);
	}
	dbg_packet_free(&pack);
	return (ret_val > 0) ? S_OK : E_FAIL;
}


HRESULT ThreadBase::getctxinfolist(
        /* [in] */ dbgint mod_no,
		/* [in] */ CTXINFOLIST *ctxinfolist) {
	dbg_srcctxinfo_request req;	
	dbg_packet pack;
	dbg_header_struct hdr;
	int ret_val;

	DBG_TRACE(("ThreadBase::getctxinfolist\n"));
	if (!ctxinfolist) return E_POINTER;
	
	if (!SERVER_INWAIT) return E_FAIL;
	
	ctxinfolist->clear();
	dbg_packet_new(&pack);
	req.mod_no = mod_no;
	dbg_packet_add_frame(&pack, FRAME_SRCCTXINFO, &req, sizeof(req));
	ret_val = request(&hdr, &pack);
	if (ret_val>0) {
		ctxinfolist->assign(&pack);
	}
	dbg_packet_free(&pack);
	return (ret_val > 0) ? S_OK : E_FAIL;
}


HRESULT ThreadBase::requestdocsrc(
        /* [in] */  dbgint mod_no,
        /* [out] */ std::string *mod_name,
        /* [out] */ std::string *mod_src) {
	dbg_packet pack;
	dbg_header_struct hdr;
	dbg_frame *frame;
	dbg_source_request src_req;
	dbg_source_body *src_body;	
	int ret_val, errcnt = 0;
	int fpos, fullsize = 0, siz;
	bool firsttime = true;
	char *str;
	bool ok;

	DBG_TRACE(("ThreadBase::requestdocsrc\n"));
	if (!SERVER_INWAIT) return E_FAIL;

	if (!mod_name) return E_POINTER;
	*mod_name = "";
	*mod_src = "";
	dbg_packet_new(&pack);
	src_req.mod_no = mod_no;
	fpos = 0;	
	while (true) {
		src_req.from_filepos = fpos;
		dbg_packet_clear(&pack);
		dbg_packet_add_frame(&pack, FRAME_SOURCE, &src_req, sizeof(src_req));
		ret_val = request(&hdr, &pack);
		frame = (ret_val > 0)? dbg_packet_findfirstframe(&pack,FRAME_SOURCE) : NULL;
		ok = (frame!=NULL);
		if (ok) {
			src_body = FRAME_DATA_PTR(dbg_source_body,frame);
			if (firsttime) {
				firsttime = false;
				if (dbg_packet_findstring(&pack, src_body->imod_name, &str, &siz)) {
					mod_name->assign(str,siz);
				}
				fullsize = src_body->full_size;
			}
			str = NULL;
			siz = 0;
			ok = (dbg_packet_findstring(&pack, src_body->itext, &str, &siz)!=0);
			if (ok) {
				mod_src->append(str,siz);
				errcnt = 0;
				fpos += siz;
				if (fpos >= fullsize) break;
			}
		}
		if (!ok && (++errcnt > 3)) break;
	}
	
	dbg_packet_free(&pack);

	if (!ok) {
		*mod_src="";
		*mod_name="";
		return E_FAIL;
	}
	return S_OK;
}

HRESULT ThreadBase::requestdocparent( 
        /* [in] */ dbgint mod_no,
        /* [out] */ dbgint *parent_mod_no) {
	return E_FAIL;
}

HRESULT ThreadBase::requestdocfilename( 
        /* [in] */ dbgint mod_no,
        /* [out] */ std::string *mod_name) {
	HRESULT hr;
	DOCLIST::iterator it;

	DBG_TRACE(("ThreadBase::requestdocfilename\n"));

	hr=getdoclist(&fdoclist);
	if (FAILED(hr)) return hr;
	if (!mod_name) return E_POINTER;

	*mod_name = "";

	hr = E_FAIL;
	for (it = fdoclist.begin(); it!=fdoclist.end(); it++) {
		if ((*it).mod_no == mod_no) {
			*mod_name = (*it).mod_name;
			hr = S_OK;
			break;
		}
	}
	return hr;
}

HRESULT ThreadBase::evaluate(
	/* [in] */ std::string &expr,
	/* [in] */ dbgint scope_id,
	/* [out]*/ std::string *serial_str,
	/* [out]*/ std::string *rslt_error) {
	dbg_packet pack;

	dbg_header_struct hdr;
	dbg_eval_request eval;
	dbg_frame *frame;
	dbg_eval_body *evalbody;
	char *rslt_str, *str;
	int ret_val, siz;
	bool ok = false, err = false;

	DBG_TRACE(("ThreadBase::evaluate()\n"));
	if (!serial_str) return E_POINTER;
	*serial_str = "";
	if (!SERVER_INWAIT) return E_FAIL;

    if (rslt_error) *rslt_error="";

	dbg_packet_new(&pack);
	eval.scope_id = scope_id;
	eval.istr = dbg_packet_add_string(&pack, expr.c_str());
	dbg_packet_add_frame(&pack, FRAME_EVAL, &eval, sizeof(eval));

	ret_val = request(&hdr, &pack);
	if (ret_val>0) {
		frame = dbg_packet_findfirstframe(&pack, FRAME_EVAL);
		if (frame == NULL) {
			frame = dbg_packet_findfirstframe(&pack, FRAME_LOG);
			if (frame != NULL) {
				onlog(&pack);
			}
		} else while (frame) {
			evalbody = FRAME_DATA_PTR(dbg_eval_body,frame);
			if ((evalbody->istr == 0 && expr.length() == 0) ||
			    (dbg_packet_findstring(&pack, evalbody->istr, &str, &siz)  && strcmp(str,expr.c_str())==0)) {

				ok = true;

				err = evalbody->ierror!=0;
				if (err &&
					dbg_packet_findstring(&pack, evalbody->ierror, &rslt_str, &siz) && siz > 0) {
					if (rslt_error==NULL) {
						ok=false;
					} else {
						rslt_error->assign(rslt_str, siz);
					}
				}
				
				if (dbg_packet_findstring(&pack, evalbody->iresult, &rslt_str, &siz) && siz > 0) {
					serial_str->assign(rslt_str, siz);
				}
				break;
			}
			frame = dbg_packet_findnextframe(&pack, FRAME_EVAL, frame);
		}
	} else if (ret_val < 0) {
		ok = true;
		err = true;
		rslt_error->assign("Debug Session has been terminated");
	}
	dbg_packet_free(&pack);

	fdoclist.clear(); // doc list might changed upon eval "include 'xxx.xxx'"

	return (ok) ? ((err) ? S_FALSE : S_OK):E_FAIL;
}

HRESULT ThreadBase::setupdatebreakpoint(
		/* [in] */ dbgint bp_no_in,
		/* [in] */ dbgint line_no,
		/* [in] */ dbgint mod_no,
		/* [in] */ const std::string &mod_name,
		/* [in] */ const std::string &condition,
		/* [in] */ bool istemp,
		/* [in] */ dbgint skiphits,
		/* [in] */ BPSTATE state,
		/* [out] */ dbgint *bp_no_out) {
	dbg_packet pack;
	dbg_header_struct hdr;
	int ret_val;
	dbg_bps_request bps;
	dbg_bpl_body *body;
	HRESULT hr = E_FAIL;
	dbg_frame *frame;
	char *ret_mod_name, *ret_condition;
	int sz;

	if (!bp_no_out) return E_POINTER;
	if (!SERVER_INWAIT) return E_FAIL;

	*bp_no_out = 0;

	dbg_packet_new(&pack);
	memset(&bps, 0, sizeof(bps));

	bps.bp_no = bp_no_in;
	bps.line_no = line_no;
	bps.mod_no = mod_no;
	bps.skiphits = skiphits;
	bps.imod_name = dbg_packet_add_string(&pack, mod_name.c_str());
	bps.icondition = dbg_packet_add_string(&pack, condition.c_str());
	bps.istemp = istemp;
	bps.state = state;

	dbg_packet_add_frame(&pack, FRAME_BPS, &bps, sizeof(bps));
	ret_val = request(&hdr, &pack);
	if (ret_val) {
		frame = dbg_packet_findfirstframe(&pack, FRAME_BPL);
		if (frame!=NULL) {
			body = FRAME_DATA_PTR(dbg_bpl_body, frame);
			dbg_packet_findstring(&pack, body->imod_name, &ret_mod_name, &sz);
			dbg_packet_findstring(&pack, body->icondition, &ret_condition, &sz);
			if (!ret_mod_name) ret_mod_name="";
			if (!ret_condition) ret_condition="";
			
			if (body->bp_no) {
				breakpointlist.update(body->bp_no, body->line_no, body->mod_no, 
					ret_mod_name, ret_condition, body->istemp!=0, body->skiphits, body->hitcount,
					(BPSTATE) body->state, body->isunderhit!=0);
			}
			*bp_no_out = body->bp_no;
			hr = (body->bp_no > 0) ? S_OK : S_FALSE;
		}
	}
	dbg_packet_free(&pack);
	
	return hr;
}

HRESULT ThreadBase::setbreakpoint(
		/* [in] */ dbgint line_no,
		/* [in] */ dbgint mod_no,
		/* [in] */ const std::string &mod_name,
		/* [in] */ const std::string &condition,
		/* [in] */ bool istemp,
		/* [in] */ dbgint skiphits,
		/* [out] */ dbgint *bp_no) {
	return setupdatebreakpoint(0, line_no, mod_no, mod_name, condition, istemp, skiphits, BPS_ENABLED, bp_no);
}

HRESULT ThreadBase::updatebreakpoint(
		/* [in] */ dbgint bp_no,
		/* [in] */ dbgint line_no,
		/* [in] */ const std::string &condition,
		/* [in] */ dbgint skiphits,
		/* [in] */ BPSTATE state) {
	dbgint bp_no_out;

	return setupdatebreakpoint(bp_no, line_no, 0, "", condition, false, skiphits, state, &bp_no_out);
}

HRESULT ThreadBase::findbreakpoint(
		/* [in] */ dbgint line_no,
		/* [in] */ dbgint mod_no,
		/* [in] */ const std::string &mod_name,
		/* [in] */ const std::string &condition,
		/* [out] */ dbgint *bp_no) {

	HRESULT hr;
	BREAKPOINTLIST::iterator it;
	std::string amod_name;

	if (!bp_no) return E_POINTER;
	*bp_no = 0;

	hr = getbreakpointlist(NULL);
	if (FAILED(hr)) return hr;
	hr = E_FAIL;
	for (it = breakpointlist.begin(); it != breakpointlist.end(); it++) {
		if (it->state != BPS_DELETED &&
			it->line_no == line_no &&
			((mod_no!=0 && it->mod_no == mod_no) || (mod_name == it->mod_name)) &&
			condition == it->condition) {
			*bp_no = it->bp_no;
			hr = S_OK;
			break;
		}
	}
	return hr;
}

HRESULT ThreadBase::getbreakpoint(
	/* [in] */ dbgint bp_no,
	/* [out] */ dbgint *line_no,
	/* [out] */ dbgint *mod_no,
	/* [out] */ std::string &mod_name,
	/* [out] */ std::string &condition,
	/* [out] */ bool *istemp,
	/* [out] */ dbgint *skiphits,
	/* [out] */ dbgint *hitcount,
	/* [out] */ BPSTATE *state,
	/* [out] */ bool *isunderhit) {
	
	HRESULT hr;
	BREAKPOINTLIST::iterator it;

	if (!line_no || !mod_no || !skiphits || !hitcount || !istemp || !state || !isunderhit) return E_POINTER;

	hr = getbreakpointlist(NULL);
	if (FAILED(hr)) return hr;

	for (it = breakpointlist.begin(); it != breakpointlist.end(); it++) {
		if (it->bp_no == bp_no) {
			*line_no = it->line_no;
			*mod_no = it->mod_no;
			mod_name = it->mod_name;
			condition = it->condition;
			*istemp = it->istemp;
			*skiphits = it->skiphits;
			*hitcount = it->hitcount;
			*state = it->state;
			*isunderhit = it->isunderhit;
			return S_OK;
		}
	}
	return E_FAIL;	
}

HRESULT ThreadBase::deletebreakpoint(dbgint bp_no) {
	dbgint bp_no_out;
	return setupdatebreakpoint(bp_no, 0, 0, "", "", false, 0, BPS_DELETED, &bp_no_out);
}

HRESULT ThreadBase::getbreakpointlist(
	/* [in] */ BREAKPOINTLIST *bplist) {

	dbg_packet pack;
	dbg_header_struct hdr;
	int ret_val;
	dbg_bpl_request req;

	DBG_TRACE(("ThreadBase::getbreakpointlist\n"));

 	if (!is_valid_bplist) {
		if (!SERVER_INWAIT) return E_FAIL;
		dbg_packet_new(&pack);
		req.bp_no = 0;
		dbg_packet_add_frame(&pack, FRAME_BPL, &req, sizeof(req));

		breakpointlist.clear();
		ret_val = request(&hdr, &pack);
		if (ret_val>0) {
			breakpointlist.assign(&pack);
		}
		dbg_packet_free(&pack);
		is_valid_bplist = 1;
	}

	if (bplist && bplist != &breakpointlist) {
		bplist->assign(breakpointlist);
	}

	return S_OK;
}

HRESULT ThreadBase::getscopelist(
	/* [in] */ SCOPELIST *ascopelist) {
	if (!ascopelist) return E_POINTER;
	ascopelist->assign(scopelist);
	return S_OK;
}

HRESULT ThreadBase::getcurloc(
	/* [out] */ dbgint *mod_no,
	/* [out] */ dbgint *line_no,
	/* [out] */ std::string *scope_name) {
	if (scopelist.size() == 0) return E_FAIL;
	if (!line_no || !mod_no || !scope_name) return E_POINTER;
	*mod_no = scopelist[0].mod_no;
	*line_no = scopelist[0].line_no;
	*scope_name = scopelist[0].scope_descr;
	return S_OK;
}

HRESULT ThreadBase::getperffreq(
		/* [out] */dbgint64 *perffreq,
		/* [out] */dbgint *perfaccuracy) {
	dbg_packet pack;
	dbg_header_struct hdr;	
	dbg_prof_c_request body;	
	int ret_val;
	dbg_frame *p;

	DBG_TRACE(("ThreadBase::getperffreq\n"));

	if (!perffreq || !perfaccuracy) return E_FAIL;
	*perffreq = 0;
	*perfaccuracy = 0;
	if (!SERVER_INWAIT) return E_FAIL;

	dbg_packet_new(&pack);
	body.test_loops = 1000;
	dbg_packet_add_frame(&pack, FRAME_PROF_C, &body, sizeof(body));

	ret_val = request(&hdr, &pack);
	if (ret_val>0) {
		p = dbg_packet_findfirstframe(&pack, FRAME_PROF_C);
		if (p) {
			dbg_prof_c_body *body;
			body = FRAME_DATA_PTR(dbg_prof_c_body,p);
			*perffreq = (unsigned dbgint)body->tm_freq_lo + (((unsigned dbgint64)body->tm_freq_hi) << 32);
			*perfaccuracy = body->tm_diff_m;
		}
	}

	dbg_packet_free(&pack);
	return (ret_val > 0) ? S_OK : E_FAIL;
}

HRESULT ThreadBase::getprofdatalist(
        /* [in] */ dbgint mod_no,
		/* [in] */ PROFDATALIST *aprofdatalist) {
	dbg_packet pack;
	dbg_header_struct hdr;
	dbg_prof_request body;
	int ret_val;

	DBG_TRACE(("ThreadBase::getprofdatalist\n"));

	if (!aprofdatalist) return E_FAIL;
	aprofdatalist->clear();
	if (!SERVER_INWAIT) return E_FAIL;
	dbg_packet_new(&pack);
	body.mod_no = mod_no;
	dbg_packet_add_frame(&pack, FRAME_PROF, &body, sizeof(body));

	ret_val = request(&hdr, &pack);
	if (ret_val>0) {
		aprofdatalist->assign(&pack);
	}
	dbg_packet_free(&pack);
	return (ret_val > 0) ? S_OK : E_FAIL;
}


HRESULT ThreadBase::setsessionoptions(
		/* [in] */ dbgint opt_flags) {
	int ret_val;
	dbg_packet send_pack;
	dbg_header_struct hdr;
	dbg_set_options_request opt_body;

	dbg_packet_new(&send_pack);
	opt_body.opt_flags = opt_flags;
	dbg_packet_add_frame(&send_pack, FRAME_SET_OPT, &opt_body, sizeof(opt_body));
	ret_val = request(&hdr, &send_pack);
	dbg_packet_free(&send_pack);
	return (ret_val>0) ? S_OK : E_FAIL;
}

