/***************************************************************************
                          ThreadBase.cpp  -  description
                             -------------------
    begin                : Fri Mar 2 2001
    copyright            : (C) 2001 by Dmitri Dmitrienko
                         : (C) 2002, 2007 NuSphere Corp.
    www                  : http://dd.cron.ru
                         : http://www.nusphere.com/
    license              : This source file is subject to version 1.01 of 
                           the License,  that is bundled with this package 
                           in the file LICENSE, and is available at through 
                           the world-wide-web at http://dd.cron.ru/license
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef WIN32
#include "config.w32.h"
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_GNUREGEX_H
#include <gnuregex.h>
#elif defined(HAVE_REGEX_H)
#include <regex.h>
#else
#ifdef HAVE_PCRE_H
#include <pcre.h>
#endif
#ifdef HAVE_PCREPOSIX_H
#include <pcreposix.h>
#endif
#endif 

#include "UrlParse.h"

#define IS_PCRE_RESULT(srclen, match, idx) \
	match[idx].rm_so != -1 && match[idx].rm_so <= srclen

#define GET_PCRE_RESULT(v, src, srclen, match, idx)								\
	if (IS_PCRE_RESULT(srclen, match, idx)) {									\
		v = src.substr(match[idx].rm_so, match[idx].rm_eo - match[idx].rm_so);	\
	}

UrlParse::UrlParse(const std::string& url) : ferr(0), fport(0) {
	regex_t re;
	regmatch_t match[11];
	int len = url.length();

	ferr = regcomp(&re, "^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\\?([^#]*))?(#(.*))?", REG_EXTENDED);
	if (ferr) return;
	ferr = regexec(&re, url.c_str(), 10, match, 0);
	regfree(&re);
	if (ferr) 
		return;

	/* no processing necessary on the scheme */
	GET_PCRE_RESULT(fscheme, url, len, match, 2);
	/* the path to the resource */
	GET_PCRE_RESULT(fpath, url, len, match, 5);
	/* the query part */
	GET_PCRE_RESULT(fquery, url, len, match, 7);
	/* the fragment */
	GET_PCRE_RESULT(ffragment, url, len, match, 9);
	/* extract the username, pass, and port from the hostname */
	if (IS_PCRE_RESULT(len, match, 4)) {

		/* extract username:pass@host:port from regex results */
		std::string strhost;		
		regex_t hostre;
		regmatch_t hostmatch[11];

		GET_PCRE_RESULT(strhost, url, len, match, 4);
		int hostlen = strhost.length();
		
		ferr = regcomp(&hostre, "^(([^@:]+)(:([^@:]+))?@)?((\\[([^]]+)\\])|([^:@]+))(:([^:@]+))?", REG_EXTENDED);
		if (!ferr) {
			ferr = regexec(&hostre, strhost.c_str(), 11, hostmatch, 0);
			regfree(&hostre);
			if (!ferr) {
				GET_PCRE_RESULT(fuser, strhost, hostlen, hostmatch, 2);
				GET_PCRE_RESULT(fpass, strhost, hostlen, hostmatch, 4);

				if (IS_PCRE_RESULT(hostlen, hostmatch, 7)) {
					GET_PCRE_RESULT(fhost, strhost, hostlen, hostmatch, 7);
				} else if (IS_PCRE_RESULT(hostlen, hostmatch, 8)) {
					GET_PCRE_RESULT(fhost, strhost, hostlen, hostmatch, 8);
				}

				std::string strport;
				GET_PCRE_RESULT(strport, strhost, hostlen, hostmatch, 10);
				fport = atol(strport.c_str());
			}
		}
	}
}
