/***************************************************************************
                          dbgbase_globals.cpp  -  description
                             -------------------
    begin                : Fri Mar 2 2001
    copyright            : (C) 2001 by Dmitri Dmitrienko
                         : (C) 2002, 2007 NuSphere Corp.
    www                  : http://dd.cron.ru
                         : http://www.nusphere.com/
    license              : This source file is subject to version 1.01 of 
                           the License,  that is bundled with this package 
                           in the file LICENSE, and is available at through 
                           the world-wide-web at http://dd.cron.ru/license
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef WIN32
#include "config.w32.h"
#endif

#ifdef WIN32
#include <Windows.h>
#else
#include <errno.h>
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#if HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "dbgbase/dbgbase_globals.h"


#ifdef _TRACE

#if HAVE_IOSTREAM
#include <iostream>
#endif

logger_cb_t tracelogger = NULL;
#endif

int oserrno(int isnetwork) {
#ifdef WIN32
	if (isnetwork)
		return WSAGetLastError();
	else
		return GetLastError();
#else
	return errno;
#endif
}

void oserrnostr(std::string& buf, int err) {

	buf = "";
	if (err == -1) {
		err = oserrno(0);
		if (!err) err = oserrno(1);
	}
	if (!err) return;
#ifdef WIN32
	{
		int i;
		char *p;
		FormatMessage(
			FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_ALLOCATE_BUFFER,
			NULL,
			err,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), /* Default language */
			(char *)&p,
			0,
			NULL
		);
		i=strlen(p);
		while (--i > 0 && p[i] < ' ');

		buf = p;
		buf.erase(i+1,buf.size());
		LocalFree(p);
	}
#else
	{
		buf = strerror(err);
	}
#endif

}

int __cdecl std_sprintf(std::string& buf, const char *format ...) {
	va_list args;
	int sz, ret_val, i;
	char *cbuf = NULL;

	va_start(args, format);
	for (sz=512, i=0; i < 16; i++, sz=sz*2) {
		cbuf = new char[sz];
#ifdef WIN32
		ret_val = vsnprintf(cbuf, sz, format, args);
#else
		{
			va_list args_copy;
			va_copy(args_copy, args);
			ret_val = vsnprintf(cbuf, sz, format, args_copy);
			va_end(args_copy);
		}
#endif

		if (ret_val >= 0 && ret_val < sz) break;
		delete cbuf;
		cbuf = NULL;
	}
	if (cbuf) {
		buf = cbuf;
		delete cbuf;
	} else {
		buf = "";
	}
	va_end(args);
	return buf.length();
}

void std_toupper(const std::string& str, std::string& buf) {
	buf = str;
	for (std::string::iterator it = buf.begin(); it!=buf.end(); it++) {
		*it = toupper(*it);
	}
}

bool fileexists(const std::string& fname) {
	char buf[PATH_MAX];
	FILE *f;
	buf[0] = 0;
	realpath(fname.c_str(), buf);
	buf[PATH_MAX-1] = 0;
	if (strlen(buf) == 0) return false;
	// php.exe generates AV on attempt to run a file that does not exist.
	f = fopen(buf, "r");
	if (f == NULL) return false;
	fclose(f);
	return true;
}

bool readfile(const std::string& fname, std::string& text) {
	char buf[PATH_MAX];
	FILE *f;
	char *src;
	dbgint len;
	
	text = "";
	buf[0] = 0;
	realpath(fname.c_str(), buf);
	buf[PATH_MAX-1] = 0;
	if (strlen(buf) == 0) return false;
	f = fopen(buf, "r");
	if (f == NULL) return false;
	fseek(f, 0, SEEK_END);
	len = ftell(f);
	src = (char*)malloc(len);
	fseek(f, 0, SEEK_SET);
	if (src) {
	    fread(src, 1, len, f);
	}
	fclose(f);
	if (src) {
	    text.assign(src, len);
	    free(src);
	    return true;
	}
	return false;
}

bool findfile(std::string& fname) {
	if (fname.find("\\") != fname.npos || fname.find("/") != fname.npos ) 
		return fileexists(fname);
	std::string origfname = fname;
	if (fileexists(fname)) return true;
	std::string pathlist(getenv("PATH"));
	while (pathlist.length() > 0) {
		int p = pathlist.find(PATH_SEP);
		if (p < 0) p = pathlist.length();
		std::string path = pathlist.substr(0, p);
		fname = path + LOCAL_FS_SLASH + origfname;
		if (fileexists(fname)) return true;
		pathlist.erase(0, p+1);
	}
	fname = origfname;
	return false;
}

#ifdef WIN32
int gettimeofday(struct timeval *ti) {
	SYSTEMTIME systime;
	GetSystemTime(&systime);	
	ti->tv_sec = ((int)systime.wHour * 60 + (int)systime.wMinute) * 60 + (int)systime.wSecond;
	ti->tv_usec = (int)systime.wMilliseconds * 1000;
	return 0;
}

char *realpath(const char *fname, char *realname) {
	return  (GetFullPathName(fname, MAX_PATH, realname, NULL) > 0) ? (realname) : (NULL);
}
#endif

bool realpath(const std::string &fname, std::string &realname) {
	char buf[PATH_MAX];
	if (realpath(fname.c_str(), buf)) {
		realname = buf;
		return true;
	}
	realname = "";
	return false;
}

int extractpath(std::string const &fname, std::string &path) {	
	int i = fname.rfind(LOCAL_FS_SLASH);
	if (i == std::string::npos) {
	    i = 0;
	}
	path.assign(fname.substr(0, i));
	return i;
}

///////////////// HELPER CLASSES ////////////////

// SCOPELIST

void SCOPELIST::assign(dbg_packet *pack) {
	dbg_frame *p;
	char *str;
	scopelistitem stkitem;
	dbg_stack_body *stk;
	int siz;

	clear();
	p = dbg_packet_findfirstframe(pack, FRAME_STACK);
	while (p) {
		stk = FRAME_DATA_PTR(dbg_stack_body,p);
		
		stkitem.line_no = stk->line_no;
		stkitem.mod_no = stk->mod_no;
		stkitem.scope_id = stk->scope_id;
		if (dbg_packet_findstring(pack, stk->idescr, &str, &siz)) {
			stkitem.scope_descr = str;
		} else {
			stkitem.scope_descr = "";
		}
		push_back(stkitem);
		p = dbg_packet_findnextframe(pack, FRAME_STACK, p);
	}
}


// BREAKPOINTLIST

void BREAKPOINTLIST::assign(dbg_packet *pack) {
	dbg_bpl_body *body;
	char *mod_name, *condition;
	int siz;
	breakpointlistitem bpitem;
	dbg_frame *frame;

	clear();
	frame = dbg_packet_findfirstframe(pack, FRAME_BPL);
	while (frame) {
		body = FRAME_DATA_PTR(dbg_bpl_body, frame);
		mod_name = "";
		condition = "";
		bpitem.bp_no = body->bp_no;
		bpitem.line_no = body->line_no;
		bpitem.mod_no = body->mod_no;
		dbg_packet_findstring(pack, body->imod_name, &mod_name, &siz);
		bpitem.mod_name = (mod_name)?(mod_name):("");
		dbg_packet_findstring(pack, body->icondition, &condition, &siz);
		bpitem.condition = (condition)?(condition):("");
		bpitem.istemp = (body->istemp!=0);
		bpitem.skiphits = body->skiphits;
		bpitem.hitcount = body->hitcount;
		bpitem.state = (BPSTATE)body->state;
		bpitem.isunderhit = (body->isunderhit != 0);
		push_back( bpitem);
		frame = dbg_packet_findnextframe(pack, FRAME_BPL, frame);
	}
}

void BREAKPOINTLIST::add(
		dbgint bp_no,
		dbgint line_no,
		dbgint mod_no,
		char *mod_name,
		char *condition,
		bool istemp,
		dbgint skiphits,
		dbgint hitcount,
		BPSTATE state,
		bool isunderhit) {

	breakpointlistitem bpitem;
	
	bpitem.bp_no = bp_no;
	bpitem.line_no = line_no;
	bpitem.mod_no = mod_no;
	bpitem.mod_name = mod_name;
	bpitem.condition = condition;
	bpitem.istemp = istemp;
	bpitem.skiphits = skiphits;
	bpitem.hitcount = hitcount;
	bpitem.state = state;
	bpitem.isunderhit = isunderhit;
	push_back(bpitem);
}

void BREAKPOINTLIST::update(
		dbgint bp_no,
		dbgint line_no,
		dbgint mod_no,
		char *mod_name,
		char *condition,
		bool istemp,
		dbgint skiphits,
		dbgint hitcount,
		BPSTATE state,
		bool isunderhit) {
	iterator it;
	for (it=begin(); it!=end(); it++) {
		if (it->bp_no==bp_no) {
			if ((state & ~BPS_UNRESOLVED) == BPS_DELETED) {
				erase(it);
			} else {
				it->line_no = line_no;
				it->mod_no = mod_no;
				it->mod_name = mod_name;
				it->condition = condition;
				it->istemp = istemp;
				it->skiphits = skiphits;
				it->hitcount = hitcount;
				it->state = state;
				it->isunderhit = isunderhit;
			}
			return;
		}
	}
	add(bp_no, line_no, mod_no, mod_name, condition, istemp, skiphits, hitcount, state, isunderhit);
}


// DOCLIST

void DOCLIST::assign(dbg_packet *pack) {
	dbg_src_tree_body *body;
	char *mod_name;
	int siz;
	doclistitem item;
	dbg_frame *frame;

	clear();
	frame = dbg_packet_findfirstframe(pack, FRAME_SRC_TREE);
	while (frame) {
		body = FRAME_DATA_PTR(dbg_src_tree_body,frame);
		mod_name = NULL;
		dbg_packet_findstring(pack, body->imod_name, &mod_name, &siz);
		if (mod_name)
			item.mod_name = mod_name;
		item.mod_no = body->mod_no;
		push_back(item);
		frame = dbg_packet_findnextframe(pack, FRAME_SRC_TREE, frame);
	}
}


// LINESINFOLIST

void LINESINFOLIST::assign(dbg_packet *pack) {
	dbg_srclinesinfo_body *body;
	linesinfoitem item;
	dbg_frame *frame;

	clear();
	frame = dbg_packet_findfirstframe(pack, FRAME_SRCLINESINFO);
	while (frame) {
		body = FRAME_DATA_PTR(dbg_srclinesinfo_body,frame);

		item.mod_no = body->mod_no;
		item.start_line_no = body->start_line_no;
		item.lines_count = body->lines_count;
		item.ctx_id = body->ctx_id;
		
		push_back(item);
		frame = dbg_packet_findnextframe(pack, FRAME_SRCLINESINFO, frame);
	}
}
	

// CTXINFOLIST

void CTXINFOLIST::assign(dbg_packet *pack) {
	dbg_srcctxinfo_body *body;
	ctxinfoitem item;
	dbg_frame *frame;
	char *str;
	int siz;

	clear();
	frame = dbg_packet_findfirstframe(pack, FRAME_SRCCTXINFO);
	while (frame) {
		body = FRAME_DATA_PTR(dbg_srcctxinfo_body,frame);

		item.mod_no = body->mod_no;
		item.ctx_id = body->ctx_id;
		if (dbg_packet_findstring(pack, body->ifunction_name, &str, &siz)) {
			item.function_name = str;
		} else {
			item.function_name = "";
		}
		
		push_back(item);
		frame = dbg_packet_findnextframe(pack, FRAME_SRCCTXINFO, frame);
	}
}


// PROFDATALIST

void PROFDATALIST::assign(dbg_packet *pack) {
	dbg_prof_body *body;
	profdataitem item;
	dbg_frame *frame;

	clear();
	frame = dbg_packet_findfirstframe(pack, FRAME_PROF);
	while (frame) {
		body = FRAME_DATA_PTR(dbg_prof_body, frame);

		item.mod_no = body->mod_no;
		item.hit_count = body->hit_count;
		item.line_no = body->line_no;
		item.tm_max = ((dbgint64)body->tm_max_lo) + (((dbgint64)body->tm_max_hi) << 32);
		item.tm_min = ((dbgint64)body->tm_min_lo) + (((dbgint64)body->tm_min_hi) << 32);
		item.tm_sum = ((dbgint64)body->tm_sum_lo) + (((dbgint64)body->tm_sum_hi) << 32);
		push_back(item);
		frame = dbg_packet_findnextframe(pack, FRAME_PROF, frame);
	}
}

// DSOURCELIST

// filemapper

void ensure_slashes(std::string &s, bool isbackslash, bool addlastslash) {
	char sl = isbackslash?'\\':'/';
	char isl = isbackslash?'/':'\\';
	while (true) {
		int p = s.find(isl);
		if (p == s.npos) break;
		s[p] = sl;
	}
	if (addlastslash && (s.empty() || s[s.length()-1]!=sl)) s = s+sl;
}

bool filemapper::mapfilename(map_kind type, const std::string &from, std::string& to, std::string& warnstr) {
	warnstr="";
	DBG_TRACE(("filemapper::mapfilename(%d, %s)\n", (int)type, from.c_str()));
	if (!is_HTTP) {
		DBG_TRACE(("mapper:fromroot:leaving(non http mode)\n"));
		if (type == mkRemoteToURL || type == mkLocalToURL) {
			warnstr = "file mapping is not available for non-HTTP mode.";
			to = "";
			return false;
		}
		to = from;
		return true;
	}	
	std::string tmpname = from;
	std::string fromroot, tmppart;
	switch (type) {
		case mkLocalToURL: 
		case mkLocalToRemote: 
			realpath(tmpname, tmpname);
			ensure_slashes(tmpname, LOCAL_FS_SLASH=='\\', false);
			fromroot=flocal_root;
			break;
		case mkRemoteToLocal:
		case mkRemoteToURL:
			fromroot = fremote_root;
			ensure_slashes(tmpname, REMOTE_FS_SLASH=='\\', false);
			break;
		case mkURLToLocal:
		case mkURLToRemote:
			fromroot = furl_root;
			ensure_slashes(tmpname, false, false);
			break;
		default:
			fromroot="<EMPTY>";
	}
	DBG_TRACE(("mapper:fromroot:%s\n", fromroot.c_str()));
	to = "";
	tmppart.assign(tmpname, 0, fromroot.length());
	if (strcoll(fromroot.c_str(), tmppart.c_str()) == 0) {
		std::string toroot;
		char sl;
		tmpname.erase(0, fromroot.length());
		switch (type) {
			case mkLocalToRemote:
			case mkURLToRemote:
				toroot=fremote_root;
				sl=REMOTE_FS_SLASH;
				break;
			case mkLocalToURL:
			case mkRemoteToURL:
				toroot=furl_root;
				sl='/';
				break;
			case mkRemoteToLocal:
			case mkURLToLocal:
				toroot=flocal_root;
				sl=LOCAL_FS_SLASH;
				break;
			default:
				sl=LOCAL_FS_SLASH;
				toroot="";
		}
		DBG_TRACE(("mapper:toroot:%s\n", toroot.c_str()));
		to=toroot + tmpname;

		ensure_slashes(to, sl=='\\', false);
	} else {
		DBG_TRACE(("mapper:not under root (%s)\n", tmppart.c_str()));
	}

	bool ok=(to.length() > 0);
	if (!ok) {
		std::string fromn, ton;
		switch (type) {
			case mkLocalToRemote:
			case mkLocalToURL: 
				fromn="local FS file";
				break;
			case mkRemoteToLocal:
			case mkRemoteToURL: 
				fromn="remote FS file";
				break;
			case mkURLToLocal: 
			case mkURLToRemote: 
				fromn="URL";
				break;
			default:
				fromn="?";
		}
		switch (type) {
			case mkURLToLocal:
			case mkRemoteToLocal: 
				ton="local FS file";
				break;
			case mkLocalToRemote: 
			case mkURLToRemote: 
				ton="remote FS file";
				break;
			case mkLocalToURL: 
			case mkRemoteToURL: 
				ton="URL";
				break;
			default:
				ton="?";
		}
		warnstr="Can not map "+fromn+" \""+from+"\" "\
				"to any "+ton+". This file must be under "\
				"\""+fromroot+"\" directory "\
				"Check mapping settings for HTTP mode.";
	}
	return ok;
}

void filemapper::modified() {
	ensure_slashes(furl_root, false, true);
	ensure_slashes(flocal_root, LOCAL_FS_SLASH=='\\', true);
	ensure_slashes(fremote_root, REMOTE_FS_SLASH=='\\', true);
}

// eval_itemtag

eval_itemtag::~eval_itemtag() {
	if (subitems) {
		delete subitems;
		subitems=NULL;
	}
}


/*************************************************************************************
 *
 *			DEBUGGER HOST PART
 *
 *************************************************************************************/

int dbh_create_listen_socket(const char* addr, unsigned short port ){
	int sock;
	sockaddr_in saddr;
	hostent *host_info;

	DBG_TRACE(("dbh_create_listen_socket(%s, %d)\n", addr, port));
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock <= 0) return -1;

	DBG_TRACE(("socket=%d\n",sock));
	memset(&saddr, 0, sizeof(saddr));
	saddr.sin_family = AF_INET;
	if (addr && addr[0]) {
		saddr.sin_addr.s_addr = inet_addr(addr);
		if (saddr.sin_addr.s_addr == INADDR_NONE){
			host_info = gethostbyname(addr);
			if (host_info == 0) return -4;
			saddr.sin_addr.s_addr = ((in_addr *)host_info->h_addr_list[0])->s_addr;
		}
	}
	else {
		saddr.sin_addr.s_addr = INADDR_ANY;
	}
	saddr.sin_port = htons(port);
	if (bind(sock, (sockaddr *) &saddr, sizeof(saddr))) {
		DBG_TRACE(("bind() failed\n"));
		SCLOSE(sock);
		return -2;
	}
	if (listen(sock,5)) {
		DBG_TRACE(("listen() failed\n"));
		SCLOSE(sock);
		return -3;
	}
	return sock;
}

int dbh_check_new_client_connect(int listen_socket, struct sockaddr *addr, int timeoutms) {
	fd_set rset;
	int ret_val;
	struct timeval timeout;
	sockaddr laddr;
	NET_SIZE_T siz;
	
	if (listen_socket<=0) return 0;

	FD_ZERO(&rset);
	FD_SET((unsigned int)listen_socket, &rset);
	if (timeoutms >= 0) {
		timeout.tv_sec = timeoutms / 1000;
		timeout.tv_usec = (timeoutms >= 1000) ? 0 : timeoutms * 1000;
	} else {
		timeout.tv_sec = 0;
		timeout.tv_usec = 0;
	}

	ret_val = select(listen_socket + 1, &rset, NULL, NULL, &timeout);
	if (ret_val>0) {
		DBG_TRACE(("select returned=%d\n",ret_val));
		siz = sizeof(laddr);
		ret_val = accept(listen_socket, &laddr, &siz);
		DBG_TRACE(("accept returned=%d\n",ret_val));
	}
	if (ret_val <= 0) {
		return (ret_val) ? (-1) : (0);
	}
	if (addr) *addr = laddr;
	return ret_val;
}

int dbh_wait_debug_event(dbg_header_struct *hdr, int client_sock, dbg_packet *pack, int timeoutms) {
	return dbg_packet_recv(hdr, pack, client_sock, timeoutms);
}

void dbh_close_socket(int socket) {
	shutdown(socket, SD_BOTH);
	SCLOSE(socket);
}

int dbh_ack(int cmd , dbg_packet *pack, int socket, int flags) {
	return dbg_packet_send(cmd, pack, socket, flags);
}

