/***************************************************************************
                          dbgbase_globals.h  -  description
                             -------------------
    begin                : Fri Mar 2 2001
    copyright            : (C) 2001 by Dmitri Dmitrienko
                         : (C) 2002, 2007 NuSphere Corp.
    www                  : http://dd.cron.ru
                         : http://www.nusphere.com/
    license              : This source file is subject to version 1.01 of 
                           the License,  that is bundled with this package 
                           in the file LICENSE, and is available at through 
                           the world-wide-web at http://dd.cron.ru/license
 ***************************************************************************/

#ifndef _DBGBASE_GLOBALS_H_
#define _DBGBASE_GLOBALS_H_

#ifdef WIN32
#include "config.w32.h"
#include <winsock2.h>
#include <windows.h>
#endif

#include <string>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <vector>

#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif


#include "dbgbase/dbg_net.h"
#include "dbgbase/dbgbase_intf.h"


#ifdef WIN32
  #include <Winerror.h>

#else
  #ifndef _HRESULT_DEFINED
    #define _HRESULT_DEFINED
    typedef long HRESULT;
	#define S_OK							((HRESULT)0x00000000L)
	#define S_FALSE							((HRESULT)0x00000001L)
	#define E_FAIL							((HRESULT)0x80004005L)
	#define E_POINTER						((HRESULT)0x80004003L)
	#define E_HANDLE						((HRESULT)0x80070006L)
	#define SUCCEEDED(Status)				((HRESULT)(Status) >= 0)
	#define FAILED(Status)					((HRESULT)(Status) < 0)
  #endif
  #ifndef  __cdecl
    #define __cdecl
  #endif
#endif

#ifdef WIN32
/* Win 32 */
typedef __int64 llong;
#else
/* Linux / Unix */
#include <time.h>
typedef long long llong;
#endif

#ifdef WIN32
  #define	snprintf _snprintf
  #define	vsnprintf _vsnprintf
  #define   pid_t int
#else
  #define	stricmp strcasecmp
  #define 	strnicmp strncasecmp
#endif

#ifndef SOCKET_ERROR
#define SOCKET_ERROR (-1)
#endif

#ifndef SD_BOTH
#define SD_BOTH SHUT_RDWR
#endif

#ifndef INADDR_NONE
#define INADDR_NONE ((uint32_t) 0xffffffff)
#endif

#ifndef va_copy
# ifdef __va_copy
#  define va_copy(dest, src)	__va_copy((dest), (src))
# else
#  define va_copy(dest, src)	memcpy(&(dest), &(src), sizeof(va_list))
# endif
#endif


#ifdef DBG_TRACE
  typedef void (*logger_cb_t)(char *buf);
  extern logger_cb_t tracelogger;
#endif

typedef struct cfgprm_tag {
	cfgprm_tag() : timeoutms_client(30000), timeoutms_process(10000), port(0), breakonload(0), breakonfinish(0) {}
	int timeoutms_client;
	int timeoutms_process;
	unsigned short port;
	int breakonload;
	int breakonfinish;
	std::string address;
} cfgprm;

#define DEFAULT_LSTNR_PORT 7869
#define DEFAULT_LSTNR_TIMEOUT 30000 /* 30 sec */
#define DEFAULT_LSTNR_ADDRESS ""
#define DEFAULT_LSTNR_BREAKONSTART 1
#define DEFAULT_LSTNR_BREAKONFINISH 1

#define DEFAULT_REQ_PORT	10001
#define REQ_PORT_COUNT		100
#define DBGSESSVAR			"DBGSESSID"

#define ALL_ADDR_STR "<ALL ADDRESSES>"
#define LOCAL_ADDR_STR "127.0.0.1"

#define CONNBUFSIZE 8192
#define CONNERRBUFSIZE 8192

////////////////// P I P E S ///////////////////////////////////

#ifdef WIN32

typedef HANDLE			pipe_t;

#define INVALID_PIPE	INVALID_HANDLE_VALUE

#define CLOSEPIPE(a)		\
	if (a!=INVALID_PIPE) {	\
		CloseHandle(a);		\
		a = INVALID_PIPE;	\
	}

#define CLOSEHANDLE(a)		\
	if ((a)!=NULL) {		\
		CloseHandle(a);		\
		a = NULL;			\
	}


#else

typedef int	pipe_t;

#define INVALID_PIPE	-1

#define CLOSEPIPE(a)		\
	if (a!=INVALID_PIPE) {	\
		::close(a);			\
		a = INVALID_PIPE;	\
	}

#endif

////////////////// H E L P E R   C L A S S E S ///////////////////////////////////

/* List of active scopes */
typedef struct scopelistitemtag {
	dbgint scope_id;
	dbgint mod_no;
	dbgint line_no;
	std::string scope_descr;
} scopelistitem, *pscopelistitem;

typedef std::vector<scopelistitem> SCOPELISTBASE;

class SCOPELIST: public SCOPELISTBASE {
public:
	void assign(dbg_packet *pack);
	void assign(const SCOPELIST &scopelist) {
		clear();
		insert(begin(), scopelist.begin(), scopelist.end());
	}
};

/* List of breakpoints */
typedef struct breakpointlistitemtag {
	dbgint bp_no;
	dbgint line_no;
	dbgint mod_no;
	std::string mod_name;
	std::string condition;
	bool istemp;
	dbgint skiphits;
	dbgint hitcount;
	BPSTATE state;
	bool isunderhit;
} breakpointlistitem, *pbreakpointlistitem;

typedef std::vector<breakpointlistitem> BREAKPOINTLISTBASE;

class BREAKPOINTLIST: public BREAKPOINTLISTBASE {
public:
	void assign(dbg_packet *pack);
	void assign(const BREAKPOINTLIST &list) {
		clear();
		insert(begin(), list.begin(), list.end());
	}
	void add(
		dbgint bp_no,
		dbgint line_no,
		dbgint mod_no,
		char *mod_name,
		char *condition,
		bool istemp,
		dbgint skiphits,
		dbgint hitcount,
		BPSTATE state,
		bool isunderhit);
	void update(
		dbgint bp_no,
		dbgint line_no,
		dbgint mod_no,
		char *mod_name,
		char *condition,
		bool istemp,
		dbgint skiphits,
		dbgint hitcount,
		BPSTATE state,
		bool isunderhit);
};

/* List of documents */
typedef struct doclistitemtag {
	dbgint mod_no;
	std::string mod_name;
} doclistitem, *pdoclistitem;

typedef std::vector<doclistitem> DOCLISTBASE;

class DOCLIST: public DOCLISTBASE {
public:
	void assign(dbg_packet *pack);
	void assign(const DOCLIST &list) {
		clear();
		insert(begin(), list.begin(), list.end());
	}
};


/* List of linesinfo */
typedef struct linesinfoitemtag {
	dbgint mod_no;
	dbgint start_line_no;
	dbgint lines_count;
	dbgint ctx_id;
} linesinfoitem, *plinesinfoitem;

typedef std::vector<linesinfoitem> LINESINFOLISTBASE;

class LINESINFOLIST: public LINESINFOLISTBASE {
public:
	void assign(dbg_packet *pack);
	void assign(const LINESINFOLIST &li) {
		clear();
		insert(begin(), li.begin(), li.end());
	}
};


/* List of ctxinfo */
typedef struct ctxinfoitemtag {
	dbgint mod_no;
	std::string function_name;
	dbgint ctx_id;
} ctxinfoitem, *pctxinfoitem;

typedef std::vector<ctxinfoitem> CTXINFOLISTBASE;

class CTXINFOLIST: public CTXINFOLISTBASE {
public:
	void assign(dbg_packet *pack);
	void assign(const CTXINFOLIST &li) {
		clear();
		insert(begin(), li.begin(), li.end());
	}
};


/* List of profdata */
typedef struct profdataitemtag {
	dbgint mod_no;
	dbgint line_no;
	dbgint hit_count;
	dbgint64 tm_min;
	dbgint64 tm_max;
	dbgint64 tm_sum;
} profdataitem, *pprofdataitem;

typedef std::vector<profdataitem> PROFDATALISTBASE;

class PROFDATALIST: public PROFDATALISTBASE {
public:
	void assign(dbg_packet *pack);
	void assign(const PROFDATALIST &li) {
		clear();
		insert(begin(), li.begin(), li.end());
	}
};


/* List of dsource */
#ifdef HAVE_DBG_OP
typedef struct dsrcnodetag {
	dbgint op_type;
	std::string constant;
	dbgint var;
	dbgint opline_num;
	dbgint fetch_type;
	dbgint op_array_ctx_id;
	dbgint EA_var;
	dbgint EA_type;
} dsrcnode, *pdsrcnode;

typedef struct dsrcitemtag {
	dbgint ctx_id;
	dbgint idx;
	dbgint opcode;
	dsrcnode result;
	dsrcnode op1;
	dsrcnode op2;
	dbgint extended_value;
	dbgint lineno;
} dsrcitem, *pdsrcitem;

typedef std::vector<dsrcitem> DSOURCELISTBASE;

class DSOURCELIST: public DSOURCELISTBASE {
public:
	void assign(dbg_packet *pack);
	void assign(const DSOURCELIST &li) {
		clear();
		insert(begin(), li.begin(), li.end());
	}
};
#endif


/* List of eval data */
typedef enum php_typetag {
	ptUnknown, 
	ptLong, 
	ptDouble, 
	ptString,
    ptArray, 
	ptObject, 
	ptBool, 
	ptResource, 
	ptRef, 
	ptSoftRef
} php_type;

class EVALLIST;

extern char *valtypes[10];

typedef struct eval_itemtag {
	std::string name;
	std::string value;
	std::string valtype;
	php_type valtypeid;
	EVALLIST* subitems;
	int var_index;
	struct eval_itemtag *ref;
	eval_itemtag() : valtypeid(ptUnknown), subitems(NULL), var_index(0), ref(NULL) {}
	eval_itemtag(const std::string& aname, const std::string& avalue, const std::string& avaltype,
		php_type avaltypeid, EVALLIST* asubitems) : 

		name(aname), value(avalue), valtypeid(avaltypeid), subitems(asubitems),
		var_index(0), ref(NULL)
		
	{
		if (avaltype.length() > 0) {
			valtype=avaltype;
		} else {
			valtype=valtypes[avaltypeid];
		}
	}
	~eval_itemtag();
} eval_item, *peval_item;

typedef std::vector<peval_item> EVALLISTBASE;
typedef std::vector<peval_item> VARLIST;


class EVALLIST: public EVALLISTBASE {
public:	
	EVALLIST(const std::string& serial_str, const std::string& name);
	EVALLIST() {};
	eval_item *Add(VARLIST* var_list, const std::string& name, const std::string& value, 
		php_type valtypeid = ptUnknown, const std::string& valtype = "", EVALLIST* subitems=NULL) {

		peval_item it;

		it = new eval_item(name, value, valtype, valtypeid, subitems);
		push_back(it);
		if (var_list) var_list->push_back(it);
		return it;
	}
	~EVALLIST() {
		clear();
	};
	void clear() {
		for (const_iterator it=begin(); it!=end(); it++) {
			delete *it;
		}
		EVALLISTBASE::clear();
	}
};

// FILEMAPPER


typedef enum tag_map_location {
	mlLocal, 
	mlRemote, 
	mlURL,
} map_location;


typedef enum tag_map_kind {
    mkLocalToRemote, 
	mkLocalToURL,
    mkRemoteToLocal, 
	mkRemoteToURL,
    mkURLToLocal,    
	mkURLToRemote,
} map_kind;


#ifdef WIN32
#define LOCAL_FS_SLASH '\\'
#define REMOTE_FS_SLASH '\\'
#define PATH_SEP ";"
#else
#define LOCAL_FS_SLASH '/'
#define REMOTE_FS_SLASH '/'
#define PATH_SEP ":"
#endif

class filemapper {
private:
	std::string furl_root;
	std::string flocal_root;
	std::string fremote_root;
public:
	bool is_HTTP;
	filemapper(bool mode_is_HTTP, std::string aurl_root, std::string alocal_root, std::string aremote_root) : 
		furl_root(aurl_root), flocal_root(alocal_root), fremote_root(aremote_root), is_HTTP(false) {}
	bool mapfilename(map_kind type, const std::string &from, std::string& to, std::string& warnstr);
	void modified();
	void set_url_root(std::string s) {
		furl_root = s;
		modified();
	}
	void get_url_root(std::string& s) {
		s = furl_root;
	}
	void set_local_root(std::string s) {
		flocal_root = s;
		modified();
	}
	void get_local_root(std::string& s) {
		s = flocal_root;
	}
	void set_remote_root(std::string s) {
		fremote_root = s;
		modified();
	}
	void get_remote_root(std::string& s) {
		s = fremote_root;
	}
};

// Misc functions

int __cdecl std_sprintf(std::string& buf, const char *format ...);
void std_toupper(const std::string& str, std::string& buf);
bool fileexists(const std::string& fname);
bool findfile(std::string& fname);

#ifdef WIN32
int gettimeofday(struct timeval *ti);
#define GETTIMEOFDAY(ti) gettimeofday(ti)
char *realpath(const char *fname, char *realname);
#else
#define GETTIMEOFDAY(ti) gettimeofday(ti, NULL)
#endif

bool realpath(const std::string &fname, std::string &realname);
//int extractpath(const char *fname, char *path, int count);
int extractpath(std::string const &fname, std::string &path);
bool readfile(const std::string& fname, std::string& text);
void get_local_ip_list(std::string& ip);

/////////////////////////// O S - R E L A T E D ///////////////////////////
/*             OS-SPECIFIC ERROR             */
int oserrno(int isnetwork);
void oserrnostr(std::string& buf, int err=-1);


/*             LISTENER SOCKET               */
int dbh_create_listen_socket(const char* addr, unsigned short port );
int dbh_check_new_client_connect(int listen_socket, struct sockaddr *addr, int timeoutms);
int dbh_wait_debug_event(dbg_header_struct *hdr, int client_sock, dbg_packet *pack, int timeoutsec);
void dbh_close_socket(int socket);
int dbh_ack(int cmd , dbg_packet *pack, int socket, int flags);

#endif
